import re
import sys
import cProfile
import regexp


def proccount(numproc, proc):
    
    # Couting the number of unique processnames in the log file

    if proc not in numproc:
	numproc[proc] = 1.0
    else:
	numproc[proc] += 1.0


def markovtransition2(transition, last, new):

    # Counting the number of transitions with two-sets

    key = last + "," + new

    if key not in transition:
	transition[key] = 1.0
    else:
	transition[key] += 1.0



def markovtransition3(transition, last, middle, new):
    
    # Transitions with three-sets.

    key = last + "," + middle + "," + new

    if key not in transition:
	transition[key] = 1.0
    else:
	transition[key] += 1.0

    

def markovstat3(transition, numproc):
    
    # Calculating the markov statistics
    
    for key1 in numproc.iterkeys():
	for key2 in numproc.iterkeys():
	    for key3 in numproc.iterkeys():
	    
		proctrans = key1 + "," + key2 + "," + key3
		   
		if proctrans in transition:
		    transition[proctrans] = transition[proctrans] / numproc[key1]


def markovstat2(transition, numproc):
    
    # Calculating the markov statistics
    
    for key1 in numproc.iterkeys():
	for key2 in numproc.iterkeys():
	    
	    proctrans = key1 + "," + key2
		   
	    if proctrans in transition:
		transition[proctrans] = transition[proctrans] / numproc[key1]



def verify(transtion, numproc):
    
    # Verify the markov statistics. Score should be close to 1.0.

    for key1 in numproc.iterkeys():
	score = 0
	for key2 in numproc.iterkeys():
	    proctrans = key1 + "," + key2
	    if proctrans in transition:
		score += transition[proctrans]


def readstat(file, oldstat):
    
    # Read the old statistic from the statistics file.

    s = re.compile('[\s]+')
   
    for line in file:
	proctrans = s.split(line)
	
	key = proctrans[0]
	value = proctrans[2]

	oldstat[key] = value

    return oldstat
	

def compareupdate(transition, oldstat, anomalylist, threshold):

    # Compare today statistics with yesterday's

    for key in transition.iterkeys():
	if key in oldstat:
	    threslow = float(oldstat[key]) * (1 - threshold)
	    threshigh = float(oldstat[key]) * (1 + threshold)
	    if (transition[key] > threshigh) or (transition[key] < threslow):
		print "The transition " + key + " differ with more than " + str(threshold*100) + "% from the previous run. Not updating the statistics file with it, this is an anomaly!"
		transition[key] = oldstat[key]
		anomalylist.append(key)

	else:
	    print "The transition " + key + " is new to us.."
	    #Legge til i anomalylist? Faa den til aa legge til transisjonen i lista.
	    	
    return transition,anomalylist


def listanomaly(log, anomalylist):
    
    print "Anomal log lines: \n"

    for item in anomalylist:
	proclist = re.split(',', item)

	prevline = log.readline()

	for line in log:
	    if (procname(prevline) == proclist[0]) and (procname(line) == proclist[1]):
	    
		print prevline
		print line + "------------------"
		
		prevline = line
	    
	    else:
		prevline = line


def markov(filename):
   
    # Start the markov calculation. Need a file name.

    log = open(filename, 'r')
    last_proc = "."
    middle_proc = "."
    transition = {}
    numproc = {}
    oldstat = {}
    anomalylist = []
    threshold = 0.4
    

    for line in log:
	new_proc = regexp.get_proc(line)
	if new_proc is not None:
	    proccount(numproc, new_proc)
	    markovtransition2(transition, last_proc, new_proc)
#	    markovtransition(transition, last_proc, middle_proc, new_proc)
	    last_proc = new_proc
#	    middle_proc = new_proc

    
    markovstat2(transition, numproc)

    log.close()

#    outfilename = str(filename) + "." + "markov"
    outfilename = 'markovstat.txt'      

    anomalylist_filename = 'anomaly.txt'

    # Checking if we already have statistics from the current system

    fileexists = 0

    try:
	outfile = open(outfilename, 'r')
	fileexists = 1
	outfile.close()
    except:
	fileexists = 0
    
    if fileexists:
	print "Old file found, reading statistics and comparing...."
	outfile = open(outfilename, 'r')
	oldstat = readstat(outfile, oldstat)
	outfile.close()
	transition,anomalylist = compareupdate(transition, oldstat, anomalylist, threshold)
    else: 
	print "No statistics file found, creating one...."
    
    outfile = open(outfilename, 'w')
	
    for key,value in transition.items():
	outfile.write(key + " : " + str(value) + "\n")
    
    outfile.close()
    

    if len(anomalylist) != 0:
        log = open(filename, 'r')
#	listanomaly(log, anomalylist)
	log.close()
	
	anomalylistfile = open(anomalylist_filename, 'a+')

	for key in anomalylist:
	    anomalylistfile.write(key + "\n")
	anomalylistfile.close()

    elif fileexists:
	print "No anomalies found today...."

#markov('/home/frode/logg/syslog')
#markov('/home/frode/logfil_idi/mail_other/20090218')

if len(sys.argv) != 2:
    print "A single file name is required to use this script."
    sys.exit(1)

filename = sys.argv[1]

markov(filename)

#cProfile.run('markov(filename)')


