import re
import datetime

# This module implements some regex functionality in order to split the log lines
# There are two functions: linesplit() and date()

def linesplit(line):
    
    # linesplit() returns the host, proc, time and msg from a single log line conforming to the syslog RFC
    
    # The following is a Jenny friendly regexp - it does not work
    # p = re.compile('((?P<time>[a-z]{3}\s+[0-9]{1,2}\s\[0-9]{2}\:[0-9]{2}\:[0-9]{2})\s(?P<facpri><[a-zA-Z0-9-._]*>\s)?(?P<server>[a-zA-Z0-9-._]+)(?P<proc>[a-zA-Z0-9-._]+)\s(?P<msg>.*))')

    # The following is a Frode friendly regexp, it works.
    p = re.compile('((?P<time>\w+\s+\d+\s\d\d\:\d\d\:\d\d)\s(?P<facpri><\S+>\s)?(?P<server>\S+) )((?P<proc>\S+)\s(?P<msg>.*))')
    match = p.match(line)

    s = re.compile('[\s]+') # Split string on one or more spaces
    r = re.compile('[\[\]\(\),=+\'\"<>\:;]') # match all non-alphanumericals which can be dropped

    z = re.compile('((?P<proc>\S+)\s(?P<msg>.*))')

    try:
        proc_nopid = r.sub(' ', match.group('proc'))
        proc = (s.split(proc_nopid))[0]
        time = match.group('time')
        msg = match.group('msg')
        host = match.group('server')
    except:
        print "Found a line that does not conform to the syslog standard...."
        return None, None, None, None
    else:
        return host, proc, time, msg

def date(datetime_string):

    # date() returns a datetime_object with year, mon day, hour, min and sec for the log line.

    # TODO: Syslog does not provide information about the year. Find some way of providing it to datetime 
    # so that transitions from one year to another is handled gracefully.

    # This could probably be a lot less messy...

    monthlist = {'Jan':1, 'Feb':2, 'Mar':3, 'Apr':4, 'May':5, 'Jun':6, 'Jul':7, 'Aug':8, 'Sep':9, 'Oct':10, 'Nov':11, 'Des':12}

    p = re.compile('((?P<month>\w+)\s+(?P<day>\d+)\s(?P<hour>\d\d)\:(?P<min>\d\d)\:(?P<sec>\d\d))')

    datetime_split = p.match(datetime_string)

    year = datetime.date.today().year
    month = monthlist[datetime_split.group('month')]
    day = int(datetime_split.group('day'))
    hour = int(datetime_split.group('hour'))
    min = int(datetime_split.group('min'))
    sec = int(datetime_split.group('sec'))

    datetime_object = datetime.datetime(year, month, day, hour, min, sec)

    return datetime_object

def get_host(line):    
    host, proc, time, msg = linesplit(line)
    return host

def get_proc(line):
    host, proc, time, msg = linesplit(line)
    return proc

def get_time(line):
    host, proc, time, msg = linesplit(line)
    return time

def get_msg(line):
    host, proc, time, msg = linesplit(line)
    return msg

def get_all(line):
    return linesplit(line)


#print get_host("Apr 29 07:00:02 paragon /USR/SBIN/CRON[5931]: (root) CMD (/usr/bin/vnstatupdate)")
