import re
import sys
import cProfile
import hashlib
import pprint
import datetime
import cPickle
import time
import regexp

# Defining colours

BLACK = '\033[30m'
BLUE = '\033[34m'
GREEN = '\033[32m'
CYAN = '\033[36m'
RED = '\033[31m'
PURPLE = '\033[35m'
BROWN = '\033[33m'
LIGHT_GREY ='\033[37m'
YELLOW = '\033[1;33m'
WHITE = '\033[1;37m'
LIGHT_BLUE = '\033[1;34m'
LIGHT_GREEN = '\033[1;32m'
LIGHT_CYAN = '\033[1;36m'
LIGHT_RED = '\033[1;31m'
LIGHT_PURPLE = '\033[1;35m'
RESET = '\033[0m'

# Debug-switch to turn on a lot of print()
debug = 0


pp = pprint.PrettyPrinter(indent=4)


# Setting the confidence for a regular message to 2

confidence_value = 2

def addmessage(hostlist, host, proc, hash, time):
    
    # Adding message to the list if it is not added before.

    if host not in hostlist:
	hostlist[host] = {}

    if proc not in hostlist[host]:
	hostlist[host][proc] = {}

    if hash not in hostlist[host][proc]:
	hostlist[host][proc][hash] = []
    
    # Check if the value is added before

    duplicate = 0
    
    for occurence in hostlist[host][proc][hash]:
	if occurence == time:
	    duplicate = 1
	    break
    
    if not duplicate:
        hostlist[host][proc][hash].append(time)

def addcandidate(candidatelist, host, process, hash, delta, lastseen, confidence_value, frame):
    if host not in candidatelist:
	candidatelist[host] = {}
    
    if process not in candidatelist[host]:
	candidatelist[host][process] = {}
    
    if hash not in candidatelist[host][process]:
	candidatelist[host][process][hash] = []
	candidatelist[host][process][hash].append(delta)
	candidatelist[host][process][hash].append(lastseen)
	candidatelist[host][process][hash].append(confidence_value)
	candidatelist[host][process][hash].append(frame)
    else:
	candidatelist[host][process][hash][1] = lastseen


	if debug:
	    print "Updating record for " + str(hash)
  
def update_candidate_confidence(candidatelist, host, process, hash, confidence_value):
    candidate_confidence = candidatelist[host][process][hash][2] 
    
    if (candidate_confidence + confidence_value) < 3:
	candidate_confidence =  candidate_confidence + confidence_value
    else:
	candidate_confidence = 3
    
    if debug:
        print "Updating confidence for " + str(hash) + "to " + str(candidate_confidence)

    candidatelist[host][process][hash][2] = candidate_confidence

def findcandidates(hostlist, candidatelist, msglist):
    
    # Locates new candidates or update old ones with the last seen value
    if debug:
	print "Finding candidates...."

    for host in hostlist:
	for process in hostlist[host]:
	    for hash in hostlist[host][process]:
		if len(hostlist[host][process][hash]) > 10:
		    occurrences_list = hostlist[host][process][hash]
		    
		    # Find a better way to compute the delta since it now only depends on three values.

		    # Add a begin and stop value to handle processes that are regular only during the night for example.

		    first_occurrence = occurrences_list[0]
		    second_occurrence = occurrences_list[1]
		    third_occurrence = occurrences_list[2]

		    delta1 = (third_occurrence - first_occurrence) / 2
		    delta2 = second_occurrence - first_occurrence

		    if (delta1 == delta2) or ((delta2 - delta1) < datetime.timedelta(0, 5)) or ((delta1 - delta2) < datetime.timedelta(0, 5)):
			delta = delta2

			if delta > datetime.timedelta(0, 60):

			    delta_max = delta + datetime.timedelta(0, 5)
			    delta_min = delta - datetime.timedelta(0, 5)
		    
			    last_occurrence = third_occurrence

			    regularity_candidate = 1
			    
			    frame = occurrences_list[len(occurrences_list) -1] - first_occurrence

			    for i in range(3, len(occurrences_list)):
				log_time = occurrences_list[i]
					
				if (log_time.utctimetuple() > (last_occurrence + delta_max).utctimetuple()) or (log_time.utctimetuple() < (last_occurrence + delta_min).utctimetuple()):
				    regularity_candidate = 0
				    break
				else: 
				    last_occurrence = log_time
			    	    		    
			    if regularity_candidate: 
				addcandidate(candidatelist, host, process, hash, delta, last_occurrence, confidence_value, frame)
				if debug:
				    print "Added " + str(hash) + " as a new candidate"
				if debug:
				    print str(host) + " - " + str(process) + " - " + msglist[hash] + " - " + str(hostlist[host][process][hash])
			    else:
				if host in candidatelist:
				    if process in candidatelist[host]:
					if hash in candidatelist[host][process]:
					    if debug:
					    	print "Found a process that failed during the walk through, updating last seen..."
					    addcandidate(candidatelist, host, process, hash, delta, occurrences_list[len(occurrences_list) - 1], -1, frame)


def checklog(hostlist, candidatelist, msglist):

    # Check the log entries for regular entries to see if there are any entries that does not follow the previous
    # established rules.
    
    # Should be incorporated in findcandidates()
    
    time_error = []
    not_reported = []

    for host in candidatelist:
	for process in candidatelist[host]:
	    for hash in candidatelist[host][process]:	
		
		if debug:
    		    print hash
		    
		if hash in hostlist[host][process]:
		    occurrences_list = hostlist[host][process][hash]
		    delta = candidatelist[host][process][hash][0]
		    last_occurrence = candidatelist[host][process][hash][1]
	    
		    delta_max = delta + datetime.timedelta(0, 20)
		    delta_min = delta - datetime.timedelta(0, 20)
		    
		    frame = candidatelist[host][process][hash][3]

		    regular = 1	
		    
		    # Hack to fix those programs that are regular only in an interval.
		    # It will NOT be able to detect a possible loss of the first message
		    # if the frame matches.

		    if (last_occurrence + frame + delta) < occurrences_list[0]:
			last_occurrence = occurrences_list[0] - delta

		    for occurrence in range(0, len(occurrences_list)):
			log_time = occurrences_list[occurrence]
			if debug:
			    print str(last_occurrence) + " - " + str(delta) + " - " + str(log_time)

			if regular:
			    if (log_time.utctimetuple() > (last_occurrence + delta_max).utctimetuple()) or (log_time.utctimetuple() < (last_occurrence + delta_min).utctimetuple()):
				#print "Something fishy has happened, message " + '"' + msglist[hash] + '"' + " from process " + str(process) + " on host " + str(host) + " should have occurred at " + str(last_occurrence + delta) + ", however, it occured at " + str(log_time)
				time_error.append(str(host) + " " + str(process) + " " + msglist[hash] + " should have occurred at " + str(last_occurrence + delta) + ", however, it occured at " + str(log_time))
		    
				update_candidate_confidence(candidatelist, host, process, hash, -1)
				regular = 0
			    else:   
				last_occurrence = log_time
		    
		    if regular:
			update_candidate_confidence(candidatelist, host, process, hash, len(occurrences_list))

		else:
		    # print "Process " + '"' + process + " at " + host + '"' + " did not report " + '"' + msglist[hash] + '"' + " although it is marked as being regular!"
		    not_reported.append(str(host) + " " + str(process) + " " + msglist[hash])
		    update_candidate_confidence(candidatelist, host, process, hash, -1)


    if len(time_error) > 0:
        print YELLOW + "Irregularities in the log file( " + str(len(time_error)) + "):" + RESET
	for error in time_error:
	    print error
    
    if len(time_error) > 0 and len(not_reported) > 0:
	print LIGHT_RED + "---------------------------------------" + RESET
    if len(not_reported) > 0:
    	print YELLOW + "Messages that are marked as regular but did not appear in this log file (" + str(len(not_reported))+"):" + RESET
    	for unreported_message in not_reported:
	    print unreported_message


def remove_faulty_candidates(candidatelist):
    
    # Removes candidates that have broken the rules too many times.
    
    delete_list_hash = []
    delete_list_process = []
    delete_list_host = []

    for host in candidatelist:
	for process in candidatelist[host]:
	    for hash in candidatelist[host][process]:
		if candidatelist[host][process][hash][2] <= 0:
		    if len(candidatelist[host][process]) > 1:
			delete_list_hash.append([host, process, hash])
		    else:
			if len(candidatelist[host]) > 1:
			    delete_list_process.append([host, process])
			else:
			    delete_list_host.append(host)
			        
    for host, process, hash in delete_list_hash:
	#if debug:
    	print "Removing faulty candidate " + str(hash) + " from " + str(process) + " at " + str(host)
	del candidatelist[host][process][hash]
    for host, process in delete_list_process:
	if debug:
	    print "Removing empty process " + str(process) + " at " + str(host)
	del candidatelist[host][process]
    for host in delete_list_host:
	if debug:
	    print "Removing empty host " + str(host)
	del candidatelist[host]

def flushmsg(candidatelist, msglist):

    # Flush messages that are not in the candidate list

    candidate_messages = {}
    for host in candidatelist:
	for proc in candidatelist[host]:
	    for hash in candidatelist[host][proc]:
		if hash in msglist:
		    candidate_messages[hash] = msglist[hash]
    
    return candidate_messages

def writelist(list, filename):
    
    # Write a dictionary or list to file

    try:
        file = open(filename, 'w')
	file.write(repr(list))
	file.close()
	if debug:
	    print filename + " written..."
    except:
	if debug:
	    print "Could not write to file..."
   
def readlist(list, filename):
    
    # Read a list or dictionary from file

    fileexists = 0

    try:
        outfile = open(filename, 'r')
        fileexists = 1
        outfile.close()
    except:
        fileexists = 0

    if fileexists:
        if debug:
	    print "Old " + filename + " found, reading statistics and comparing...."
        file = open(filename, 'r')
	
	list = eval(file.read())

        file.close()

    return list


 
def main():
    if len(sys.argv) != 2:
	sys.exit()

    
    log_filename = sys.argv[1]

    logfile = open(log_filename, 'r')

    hostlist = {}
    msglist = {}
    candidatelist = {}

    candidate_filename = 'candidates.txt'
    hostlist_filename = 'hostlist.txt'
    msglist_filename = 'msgs.txt'

    msglist = readlist(msglist, msglist_filename)
    candidatelist = readlist(candidatelist, candidate_filename)

    linecount = 1

    for line in logfile: 
	if (linecount % 10000) == 0:
	    print "Processed " + str(linecount / 1000) + "k log lines...."
	
	host, proc, timestring, msg = regexp.get_all(line)

	if host is not None:
	    hash = hashlib.md5(msg).hexdigest()
	    time = regexp.date(timestring)
	    addmessage(hostlist, host, proc, hash, time)
    
	    if hash not in msglist:
		msglist[hash] = msg
	linecount += 1
   
    logfile.close()

    if len(candidatelist) > 0 and len(msglist) > 0:
	if debug:
	    print "Checking the candidates found in earlier files with the current log file"
	checklog(hostlist, candidatelist, msglist)
    
    # Generate new candidates
    findcandidates(hostlist, candidatelist, msglist)
    
    # Removing candidates that were not candidates after all

    remove_faulty_candidates(candidatelist)
    
    # Flushing the message list for messages that can not be candidates

    candidate_messages = flushmsg(candidatelist, msglist)

    #pp.pprint(candidatelist)
    #pp.pprint(candidate_messages)


    #writelist(hostlist, hostlist_filename)
    
    writelist(candidatelist, candidate_filename)
    writelist(candidate_messages, msglist_filename)

    if debug:
	writelist(hostlist, 'hostlist.txt')

    

if __name__=="__main__":
    main()
