# -*- coding: utf-8 -*-

from msg_extraction import extract_message
import re, pprint, math, numpy, sys, os, getopt
from pylab import * 

def remove_trailing_stars(list):
    msg = []
    prev = ''
    for item in list:
        if (prev == '*') and (item == '*'):
            continue
        msg.append(item)
        prev = item

    return msg

def add_to_dict(dict, item):
    if item in dict:
        val = dict[item]
        dict[item] += 1
    else:
        dict[item] = 1

def count_1_itemsets(log):
    print "Counting 1-itemsets...."
    
    f = open(log,'r')

    one_itemsets = {} 
    number_of_loglines = 0

    for ln in f:
        number_of_loglines += 1
        
        L = extract_message(ln.decode('utf-8', 'ignore')) #remove all weird characters, convert to unicode utf-8
        #L = identify_emailaddr(list)
        
        if len(L) > 0:
            for index, val in enumerate(L):
                item = str(index) + ':' +val
                
                if item in one_itemsets:
                    one_itemsets[item] += 1
                else:
                    one_itemsets[item] = 1
        else:
            print "no match: ",ln
    
    f.close()
    return one_itemsets, number_of_loglines

def generate_itemsets(log, one_itemsets):
    
    if debug: print "Generating n-itemsets..."
    itemsets = {}
    line_counter = 0
    f= open(log,'r')
    
    for ln in f:
        
        line_counter += 1
           
        L = extract_message(ln.decode('utf-8', 'ignore')) #remove all weird characters, convert to unicode utf-8
        
        keywords = []
        frequencies = []    
        
        for index, word in enumerate(L):
            key = str(index) + ':' + word
     
            c = one_itemsets[key]
            keywords.append(key)
            frequencies.append(c)
                
        itemset = remove_lower_20pc_area(frequencies, L)
        
        add_to_dict(itemsets, ' '.join(itemset))
 
    f.close()
    
    return itemsets


def remove_lower_20pc_area(frequencies, L): 
        
    #itemset = L[:] #Select all values in L
    itemset = []
    freq_median = median(frequencies)
    
    deletion_candidates = []
    deletion_candidates_area = 0
    limit = 0.2    
        
    bars = {} # freq -> [index]
    
    for index,freq in enumerate(frequencies):
        if freq not in bars:
            bars[freq] = []
        bars[freq].append(index)
    
    total_area = len(L)
    
    max_freq = max(frequencies)
    
    last_added = 0
    
    
    if len(bars.keys()) > 1:
        for bar, list in sorted(bars.iteritems()): # In incresing frequency order
            
            area = len(list)
            
            # If a smaller frequency has been added, larger frequencies can not be deleted
            if (last_added > 0) and (last_added < bar):
                break
            
            for index in sorted(list):
                
                if float(bar)/max_freq < limit:
                #if float(bar)/freq_median < limit:
                    count = deletion_candidates_area + area
                    
                    if bar in deletion_candidates:
                        continue
                    
                    if count < 0.4*total_area:
                        deletion_candidates.append(bar)
                        deletion_candidates_area += area
                    else:
                        last_added = bar                            
                else:
                    last_added = bar
    
    del_indexes = []
    
    for bar in deletion_candidates:
        for index in bars[bar]:
            del_indexes.append(index)
    
    for index, item in enumerate(L):
        if index not in del_indexes: 
            itemset.append(item)
                    
            
    #if ' '.join(itemset) == '':
    if debug:
        (histogram, bin_edges) = numpy.histogram (frequencies, bins=len(L), range=(0,max(frequencies)), normed=False)
        print "---------------"
        print L
        print "for deletion: ", del_indexes
        print "frequency: ", frequencies
        print "histogram: ", histogram
        
        print "result: ", itemset
        
        #fig = figure()
        #hist(frequencies, bins=len(L), range=(0,max(frequencies)))
        #show()
 
    return itemset

def remove_smallest_values(frequencies, L):
    
    itemset = L[:] # Working copy of L
    
    (histogram, bin_edges) = numpy.histogram (frequencies, bins=len(L), range=(0,max(frequencies)), normed=False)
    
def median(numbers):
   "Return the median of the list of numbers."
   # Sort the list and take the middle element.
   n = len(numbers)
   copy = numbers[:] # So that "numbers" keeps its original order
   copy.sort()
   if n & 1:         # There is an odd number of elements
      return copy[n // 2]
   else:
      return (copy[n // 2 - 1] + copy[n // 2]) / 2

def average(list):
    s = sum(list)
    return s/len(list)

def sum(list):
    sum = 0
    for i in list:
        sum += i
    return sum

def save_itemsets(itemset_file, itemsets):
    
    print "Writing result..."
    
    fw = open(itemset_file, 'w+')
    
    l = []
    
    for key, count in itemsets.items():
        key_list = key.split()
        key_list.append(str(count))
        key_list.append('\n')
            
        key_str = ' '.join(key_list)

        l.append(key_str)
    
    list = sorted(l)
    
    for element in list:
        try:
            fw.writelines(element)
        except:
            print "Something went wrong when writing data to file"
            pass
        
    fw.close()

def read_stored_one_itemsets(one_itemset_file):
    f = open(one_itemset_file, 'r')
    
    itemsets = {}
    
    for line in f:
        l = line.split()
        # One-itemset lines consists of one keyword and a count
        itemset = l[0]
        freq = l[1]
        # Itemsets can only occur once
        itemsets[itemset] = int(freq)
        
    return itemsets
        
    f.close()



def main() :
    
    global debug
    global logfile
    global one_itemsets_file
    global patternfile

    one_itemsets, number_of_loglines = count_1_itemsets(logfile)
    number_of_unique_items = len(one_itemsets.keys())
    save_itemsets(one_itemsets_file, one_itemsets)

    if debug: 
        print "\tUnique 1-itemsets found: ", number_of_unique_items
        print "\tNumber of loglines read: ", number_of_loglines
        print "\n\tGenerate patterns ..."
    
    one_itemsets = read_stored_one_itemsets(one_itemsets_file)
    n_itemsets = generate_itemsets(logfile, one_itemsets)
    save_itemsets(patternfile, n_itemsets)

    print "Finished"

def usage() :
    print "\nUsage of histogram-based pattern mining : "
    print "\t-h\t(--help) To show this message."
    print "\t-l\tFull path to the logfile to analyse (mandatory)."
    print "\t-p\tFull path to the directory where the results should be written (mandatory)."
    print "\t-d\t(--debug) To turn on debugging code."


if __name__ == "__main__" :

    debug = False
    logfile = None
    one_itemsets_file_name = 'patterns_one_itemsets'
    one_itemsets_file = None
    pattern_file_name = 'patterns'
    patternfile = None

    argv = sys.argv[1:]

    try:
        opts, args = getopt.getopt(argv, "hdl:p:", ["help", "debug"])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    if ( '-l' not in argv ) or ( '-p' not in argv ) :
        usage()
        sys.exit(2)

    for opt, arg in opts :
        if opt in ('-h', '--help') :
            usage()
            sys.exit()
        elif opt in ('-d', '--debug' ) :
            debug = True
        elif opt == '-l' :
            if not os.path.isfile( arg ) :
                print "Logfile '%s' does not exist." % arg
                sys.exit()
            logfile = arg
        elif opt == '-p' :
            if not os.path.isdir( arg ) :
                print "'%s' is not a valid directory." % arg
                sys.exit()
            path = arg
            if arg[-1] != '/':
                path = arg+'/'
            patternfile = path+pattern_file_name
            one_itemsets_file = path+one_itemsets_file_name   

    main() 
