# -*- coding: utf-8 -*-
from msg_extraction import extract_message
import re, pprint, os.path

def remove_trailing_stars(list):
    msg = []
    prev = ''
    for item in list:
        if (prev == '.*') and (item == '.*'):
            continue
        msg.append(item)    
        prev = item   
        
    return msg

def count_1_itemsets(f, debug=False):
    f.seek(0)
    frequent_1_itemsets = {}
    number_of_loglines = 0
    
    for line in f:
        number_of_loglines += 1 
        l = extract_message(line)
        
        if len(l) > 0:
            for index, val in enumerate(l):
                i = val.replace("'","\\'")
                item = str(index) + ':' + i   # add position
                
                if item in frequent_1_itemsets:
                    frequent_1_itemsets[item] += 1
                else:
                    frequent_1_itemsets[item] = 1
#                else:
#                    print "Word %s contains illegal characters" % item
        else:
            if debug: print "\nNo match for line: ",line
     
    # Register the total number of unique items 
    number_of_unique_items = len(frequent_1_itemsets.keys())
        
    return (frequent_1_itemsets, number_of_loglines, number_of_unique_items)
              
def prune_infrequent_1_itemsets(frequent_itemsets, threshold):
    itemsets = {}    

    for key, value in frequent_itemsets.items():
        if (value < threshold) or (len(key) < 2) or (len(key) > 60): # if 1-itemset very large, most likely a path
            continue
        itemsets[key] = value

    return itemsets            

def prune_infrequent_n_itemsets(frequent_itemsets, threshold):
    itemsets = {}
    
    for key, value in frequent_itemsets.items():
        #print size, ' : ' + key, value
        if (value < threshold):
            #outliers.append(key)
            del frequent_itemsets[key]
            
def count_n_itemsets(f, frequent_1_itemsets, debug=False):
    f.seek(0)
    frequent_n_itemsets = {} # size -> (itemset,count)

    for line in f: 
        l = extract_message(line)
        tmp_set = []
        
        for index, val in enumerate(l):
            item = str(index) + ':' +val
            if item in frequent_1_itemsets:
                tmp_set.append(val)
            else:
                tmp_set.append('.*')
        size = len(tmp_set)
        num_of_infrequent = tmp_set.count('.*')
        
#        itemset = remove_trailing_stars(tmp_set)
        itemset = tmp_set
        
        if num_of_infrequent < (size-3) : # don't count if only three frequent item
            itemsetstr =' '.join(itemset)
                
            add_to_dict(frequent_n_itemsets, itemsetstr) 
            
           
        # If no frequent words found, add entire line
        else:
            line_size = len(l)
            line = ' '.join(l)
            
            add_to_dict(frequent_n_itemsets, line)

    return frequent_n_itemsets


def add_to_dict(dict, item):
    if item in dict:
        val = dict[item]
        dict[item] += 1
    else:
        dict[item] = 1

def save_itemsets(itemset_file, itemsets):
    fw = open(itemset_file, 'w+')
    
    l = []
    
    for key, count in itemsets.items():
        key_list = key.split()
        key_list.append(str(count))
        key_list.append('\n')
            
        key_str = ' '.join(key_list)
        
        l.append(key_str)
    
    list = sorted(l)
    
    for element in list:
        fw.writelines(element)
        
    fw.close()

def itemset_gen(log, threshold, itemset_file=False, return_one_itemsets_only=False, debug=False):
    
    f = open(log, 'r')
    
    (one_itemsets, number_of_loglines, number_of_unique_items) = count_1_itemsets(f)
    
    #t = int(threshold*number_of_unique_items)
    t = int(threshold*number_of_loglines)
   
    if debug: print "Generating itemsets ..." 
    if debug: print "\nThreshold is: ", t
    
    if debug: print "\nBefore prune, 1-itemsets: ", len(one_itemsets.keys())
    frequent_1_itemsets = prune_infrequent_1_itemsets(one_itemsets,t)
    if debug: print "\nAfter prune: ", len(frequent_1_itemsets.keys())
   
    if debug and itemset_file :
        path = os.path.dirname(itemset_file)
        one_itemsets_file = path + '/patterns_1_itemsets'
        save_itemsets(one_itemsets_file, one_itemsets)
        print "\tFrequent 1-itemsets saved to %s." % ( one_itemsets_file ) 

    if return_one_itemsets_only :
        if debug: print "One-itemsets will be returned."
        return frequent_1_itemsets
    else : 
        frequent_itemsets = count_n_itemsets(f, frequent_1_itemsets)
    
    #    print "Before prune n-itemsets: ", len(frequent_n_itemsets.keys())
    #    prune_infrequent_n_itemsets(frequent_n_itemsets, t)
    #    print "After prune: ", len(frequent_n_itemsets.keys())
    
        save_itemsets(itemset_file, frequent_itemsets)
    
    f.close()
    
    
