# -*- coding: utf-8 -*-

import re

def extract_message( line ) :
    " Takes a Syslog formatted logline and extracts the message part. "
    
    # Format of a Syslog logline
    p = re.compile('([A-Z]{3}\s{1,2}[0-9]{1,2}\s[0-9]{2}\:[0-9]{2}\:[0-9]{2}\s(<[A-Z0-9-._]+>\s)?(?P<server>[A-Z0-9-._]+) )(?P<msg>.*)',re.IGNORECASE)
    match = p.match(line)
    
    if not match :
        print "Line: ", line
        raise Exception, "Fetching message part of logline failed. The logline format is not recognized by this msg extractor. "

    msg = []
    msg_part = match.group( 'msg' )
      
    msg_wo_emailaddr = msg_part
        
    if '@' in msg_part:
        # Characters allowed in email-addresses: . ! # $ % & ' * + - / = ? ^ _ ` { | } ~
        # Removing = from pattern, since often used in pattern to=emailaddr f.ex
        # Safer to keep keywords. There is a chance that splitted keywords may be merged
        # with the "original" pattern anyway.
        # Not allowing an item consisting of letters only to be followed by a = sign in front of the email
        # address may solve the most frequent problems
        email_p = re.compile("(^[A-Z]=)*[A-Z0-9.!#$%'*+\-/?^_`{|}~=]+@[A-Z0-9.-]+\.[A-Z]{2,6}", re.IGNORECASE)
        msg_wo_emailaddr = email_p.sub(' ', msg_part)
    
    # Fuzzy matching of ipv6 addr, don't care if it is not a valid ipv6
    # ipv4 addresses are matched when removing numberwords
    ipv6 = re.compile("[0-9A-F]{0,4}(:[0-9A-F]{0,4}){2,7}", re.IGNORECASE)
    msg_wo_ipv6 = ipv6.sub(' ', msg_wo_emailaddr)
          
    r = re.compile('[^A-Z0-9.\-/_+]', re.IGNORECASE) # match all non-alphanumericals which can be dropped
    msg_wo_special_characters = r.sub(' ', msg_wo_ipv6) 
    
    msg_wo_trailing_ws = msg_wo_special_characters.rstrip() #remove trailing whitespaces
    
    s = re.compile('[\s]+') # Split string on one or more spaces
    msg_list = s.split(msg_wo_trailing_ws)

    msg = remove_numberwords(msg_list)
    #msg = remove_scw(msg_wo_numberwords) 
         
    return msg

def remove_numberwords(list):
    msg = []
    n = re.compile('^[0-9.\-/+]+$') # match all words consisting of digits only
    for item in list:
        match = n.match(item)
#        if match:
#            msg.append('.*')
        if not match:
            msg.append(item)
    return msg

def remove_scw(list):
    msg = []

    for item in list:
        if len(item) > 1:
            msg.append(item)

    return msg

