# -*- coding: utf-8 -*-

from msg_extraction import extract_message
from itemset_gen import itemset_gen
import re, pprint, sys, getopt, os
import cProfile, pstats
#from pattern_operations import fuzzy_dict_search as FS
#from pattern_operations import create_patterns


# python -m cProfile myscript.py

def remove_trailing_stars(list):
    msg = []
    prev = ''
    for item in list:
        if (prev == '.*') and (item == '.*'):
            continue
        msg.append(item)    
        prev = item   
        
    return msg

def add_to_dict(dict, item):
    if item in dict:
        val = dict[item]
        dict[item] += 1
    else:
        dict[item] = 1

def median(numbers):
   "Return the median of the list of numbers."
   # Sort the list and take the middle element.
   n = len(numbers)
   copy = numbers[:] # So that "numbers" keeps its original order
   copy.sort()
   if n & 1:         # There is an odd number of elements
      return copy[n // 2]
   else:
      return (copy[n // 2 - 1] + copy[n // 2]) / 2
  
def prune_infrequent_one_itemsets(one_itemsets, threshold):
    pruned = 0
    
    for key, val in one_itemsets.items():
        if val <= threshold:
            pruned += 1
            del one_itemsets[key]
    print "Number of pruned items: ", pruned
     
#def generate_primary_patterns(file, one_itemsets, threshold):
def generate_primary_patterns(file, one_itemsets, debug=False):
    
    if debug: print "Generating temporary patterns ..."
    
    patterns = {}
    lines_read = 0
    dropped = 0
   

    f = open(file, 'r')
    
    for line in f:
        
        lines_read += 1
        L = extract_message(line.decode('utf-8', 'ignore')) #remove all weird characters, convert to unicode utf-8
        
        itemset = []
        
        for index, item in enumerate(L):
            #key = item
            key = str(index) + ':' + item
            if key not in one_itemsets :
                continue
            #if one_itemsets[key] > threshold:
            #    itemset.append(item)
            itemset.append(item)
                       
        if len(itemset) < 2: # Not interresting
            
#            print L
#            print itemset
#            for item in L:
#                print item, one_itemsets[item]
#            print '--------------'
            dropped += 1
            continue
                
        itemset_str = ' '.join(itemset)
        if itemset_str not in patterns:
            patterns[itemset_str] = []


        if ' '.join(itemset) == '':
            print L
            for i, el in enumerate(L):
                k = str(i) + ':' + el
                if k in one_itemsets:
                    print k, one_itemsets[k]
        
    f.close()
    if debug: print "\tNumber of dropped lines: ", dropped
    if debug: print "\tLines read: ", lines_read
    return patterns

def save_primary_patterns(file, primary_patterns):
    print "Saving primary patterns ..."
    
    f = open(file, 'w+')
    
    tmp_list = []
    
    for key in primary_patterns.keys():
        line = key + '\n'
        f.writelines(line)
        #tmp_list.append(line)
        
#    for elem in sorted(tmp_list):    
#        f.writelines(line)
        
    f.close()

def classify_loglines(file, primary_patterns):
    print "Classifying loglines ..."
    
    primary_patterns['outliers'] = []
    
    f = open(file, 'r')

    for line in f:
        
        L = extract_message(line.decode('utf-8', 'ignore'))
        
        category = fuzzy_dict_search(primary_patterns, L)

        if category != None:
            primary_patterns[category].append(L)
        else:
            primary_patterns['outliers'].append(L)

    f.close()

def fuzzy_dict_search(dict, list):
    
    closest_key = None
    common_items = 0
    
    for key in dict.keys():
        l = key.split()
        c = compare(l,list)
        if c > common_items:
            common_items = c
            closest_key = key
    
    if common_items < 2:
        return None
    return closest_key
              
def compare(key, list): # s1 and s2 are lists
    common = []
    prev_index = -1
    
    for item in list:
        if item in key:
            i = key.index(item)
            if i > prev_index:
                prev_index = i
                common.append(item)
            else:                    # if prev_index is larger than this index, items not in correct order
                #print "Items not in order according to the key"
                common = []
                break
    
    return len(common)

def comparison_fraction(list1, list2):  # and b are lists
   
   n, m = len(list1), len(list2)
   
   if n != m: # If lines not equally long, not same pattern
       return 0.0
   
   count = 0.0    
   
   for k in range(n):
       a,b = list1[k], list2[k]
       
       if (a != b):
           continue
       count += 1.0
   
   return (count/n)

def merge(line1, line2):
    common = []
    
    for index in range(len(line1)):
        if line1[index] == line2[index]:
            common.append(line1[index])
        else:
            common.append('*')
            
    return common

def create_patterns(patterns, lines): # lines is of type list
    
    spent_lines = []
    
    common_items_factor = 0.6
    
    for i, line1 in enumerate(lines):
        
#        TODO: 
#        Problem: When identical lines are found, max_common_items_fraction
#        is not updated, so that similar lines can be merged. 
#        This becomes a problem for short lines, when it is not necessarily desirable 
#        to merge at 0.6, but rather at 0.7 or more....
        
        max_common_items_fraction = 0
        candidates = []
        count = 1
        
        if line1 in spent_lines:
            continue
        
        for j in range(i+1,len(lines)):
            
            line2 = lines[j]
            
            if line2 in spent_lines:
                #print line2, " in spent_lines!"
                continue
            
            # If line1 and line2 are identical, don't try to make pattern based on them!
            if line1 == line2:
                count += 1
#                print "Lines are identical!"
                continue
            
#            TODO:
#            Is there any way to optimize when number of similar patterns grows large?!
#            ref pattern: mimedefang.pl MDLOG HELO/EHLO torino.idi.ntnu.no pretend to be from IDI but is not.
            
            common_items_fraction = comparison_fraction(line1,line2) 
            
            if common_items_fraction >= common_items_factor:
                
                # If number of common items is larger than what has appeard so far,
                # reset candidates and replace max number of common items.
                # Otherwise, if number of common items is the same, add new candidate
                if common_items_fraction > max_common_items_fraction:
#                    if max_common_items_fraction == 0:
#                        print "First friend found:    ", line2
#                    else:
#                        print "Better friend found:   ", line2                   
                    candidates = [line2] # Reset to only include this new line
                    max_common_items_fraction = common_items_fraction
 
                elif common_items_fraction == max_common_items_fraction:
                    candidates.append(line2)
#                    print "Another friend found: ",line2
        
        # Candidates + the line they were all checked against + possibly any identical lines
        pattern_count = len(candidates) + count 
        spent_lines.append(line1)
        tmp_pattern = line1
        pattern= []
        pattern_str = None
        
                
        if len(candidates) > 0:
                        
            for candidate in candidates:
                spent_lines.append(candidate) # line can only be part of one pattern
                tmp = merge(tmp_pattern, candidate)
                tmp_pattern = tmp
            
            # Delete all *s
            for item in tmp_pattern:
                if item != '*':
                    pattern.append(item)
        else:
            pattern = tmp_pattern            
        
            
        pattern_str = ' '.join(pattern)
        
        if pattern_str not in patterns:
            patterns[pattern_str] = pattern_count
        else:
            prev_count = patterns[pattern_str]
            patterns[pattern_str] = prev_count + pattern_count
            
def generate_patterns(primary_patterns):
    print "Generating final patterns ..."
    
    patterns = {}
    
    for tmp_pattern, lines in primary_patterns.items():
#        if tmp_pattern == "mimedefang.pl MDLOG grey and":
#            print "    ", tmp_pattern
#            for line in sorted(lines):
#                print line
#            print "****************************"
        create_patterns(patterns, lines)

    return patterns

def save_patterns(file, patterns, debug=False):
    if debug: print "Saving patterns ..."
    
    f = open(file, 'w+')
    
    list = []
    number_of_lines = 0
    single_patterns = 0
    
    for key, count in patterns.items():
        number_of_lines += count
        if count == 1:
            single_patterns += 1
        line = key + ' ' + str(count) + '\n'
        #f.writelines(line)
        list.append(line)
        
    for elem in sorted(list):
        f.writelines(elem)
        
    if debug: print "\tNumber of lines represented by patterns:  ", number_of_lines
    if debug: print "\tSingle patterns ", single_patterns
    f.close()

def usage() :
    print "\nUsage of primary sorting : "
    print "\t-h\t(--help) To show this message."
    print "\t-l\tFull path to the logfile to analyse (mandatory)."
    print "\t-p\tFull path to the directory where the results should be written (mandatory)."
    print "\t-d\t(--debug) To turn on debugging code."
    print "\t-t\t(--threshold) Threshold for when a logline item is considered frequent. Default is 0.05."

def main() :

    global debug
    global threshold
    global logfile
    global primary_patterns_file
    global patternfile

    # Generate frequent 1-itemsets
    one_itemsets = itemset_gen(logfile, threshold, return_one_itemsets_only=True, debug=debug)

    #t = 0.05
    #threshold = int(t*len(one_itemsets.keys()))
    #threshold = 1000

    if debug : 
	print "\tNumber of frequent unique items: ", len(one_itemsets.keys())
	m = 0
	n = 0
	if one_itemsets:
	    m = max(one_itemsets.values())
	    n = median(one_itemsets.values())
	print "\tMax Frequency: " ,m
	print "\tMedian: ",n
	print "\tThreshold: ", threshold

    #prune_infrequent_one_itemsets(one_itemsets, threshold)
    #print "Number of unique items, after prune: ", len(one_itemsets.keys())
    #primary_patterns = generate_primary_patterns(testlog, one_itemsets, threshold)
    
    primary_patterns = generate_primary_patterns(logfile, one_itemsets,debug=debug)

    if debug : print "\nNumber of primary patterns: ", len(primary_patterns.keys())

    if debug:
        save_primary_patterns(primary_patterns_file, primary_patterns)

    classify_loglines(logfile, primary_patterns)

    patterns = generate_patterns(primary_patterns)

    #for i,v in patterns.items():
    #    print i,v

    save_patterns(patternfile, patterns)

    print "Finished"

if __name__ == "__main__" :
    
    debug = False 
    threshold = 0.05
    logfile = None
    primary_patterns_file_name = 'patterns_primary'
    primary_patterns_file = None 
    pattern_file_name = 'patterns'
    patternfile = None

    argv = sys.argv[1:]

    try:
        opts, args = getopt.getopt(argv, "hdt:l:p:", ["help", "debug", "threshold="])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    if ( '-l' not in argv ) or ( '-p' not in argv ) :
        usage()
        sys.exit(2)

    for opt, arg in opts :
        if opt in ('-h', '--help') :
            usage()
            sys.exit()
        elif opt in ('-d', '--debug' ) :
            debug = True
        elif opt == '-l' :
            if not os.path.isfile( arg ) :
                print "Logfile '%s' does not exist." % arg
                sys.exit()
            logfile = arg
        elif opt == '-p' :
            if not os.path.isdir( arg ) :
                print "'%s' is not a valid directory." % arg
                sys.exit()
            path = arg
            if arg[-1] != '/':
                path = arg+'/'
            patternfile = path+pattern_file_name
            primary_patterns_file = path+primary_patterns_file_name
        elif opt in ('-t', '--threshold') :
            try :
                t = float(arg)
            except :
                print "Threshold must be a decimal number. Using default value %s." % threshold
                t = threshold
            threshold = t

    main()
