# -*- coding: utf-8 -*-
from msg_extraction import extract_message
import xml.etree.ElementTree as ET
from pprint import pprint
from sets import Set
import re, sys, os, getopt

# Return a list including the message words only
# Code from itertools doc on python.org
def combinations(iterable, r):
    # combinations('ABCD', 2) --> AB AC AD BC BD CD
    # combinations(range(4), 3) --> 012 013 023 123
    pool = tuple(iterable)
    n = len(pool)
    if r > n:
        return
    indices = range(r)
    yield tuple(pool[i] for i in indices)
    while True:
        for i in reversed(range(r)):
            if indices[i] != i + n - r:
                break
        else:
            return
        indices[i] += 1
        for j in range(i+1, r):
            indices[j] = indices[j-1] + 1
        yield tuple(pool[i] for i in indices)

def generate_combinations( file, debug=False ) :
    
    if debug: print "Generating item combinations ... "

    item_combinations = {}

    f = open(file, 'r')
    for line in f :
        msg = extract_message(line)
        lm = len(msg) + 1
        
        combs = (i for length in range( 1, lm ) for i in combinations( msg, length ) )
        
        for comb in combs :
            if comb not in item_combinations :
                item_combinations[comb] = 0
            item_combinations[comb] += 1

    return item_combinations

def save_results(dict) :
    
    global items_file, debug

    if debug: print "Writing results ..."
    
    tmp_list = []

    for item, count in dict.items():
        i = str(item) + " " + str(count) + "\n"
        tmp_list.append(i)
    
    list = sorted(tmp_list)
    
    fw = open(items_file, 'w')   
 
    for elem in list:
        fw.writelines(elem)
    fw.close()


def main() :

    global debug
    global logfile
    global patternfile

    item_combinations = generate_combinations( logfile, debug=debug )
   
    save_results(item_combinations)

 
    # The generation of all combinations of a loglines items is too slow to
    # handle normal loglines. No point in continuing.
   
    #pprint(item_combinations) 


    print "Finished"

def usage() :
    print "\nUsage of histogram-based pattern mining : "
    print "\t-h\t(--help) To show this message."
    print "\t-l\tFull path to the logfile to analyse (mandatory)."
    print "\t-p\tFull path to the directory where the results should be written (mandatory)."
    print "\t-d\t(--debug) To turn on debugging code."

if __name__ == "__main__" :

    debug = False
    threshold = 0.05
    logfile = None
    pattern_file_name = 'patterns'
    patternfile = None
    combinations_file = None

    argv = sys.argv[1:]

    try:
        opts, args = getopt.getopt(argv, "hdl:p:", ["help", "debug"])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    if ( '-l' not in argv ) or ( '-p' not in argv ) :
        usage()
        sys.exit(2)

    for opt, arg in opts :
        if opt in ('-h', '--help') :
            usage()
            sys.exit()
        elif opt in ('-d', '--debug' ) :
            debug = True
        elif opt == '-l' :
            if not os.path.isfile( arg ) :
                print "Logfile '%s' does not exist." % arg
                sys.exit()
            logfile = arg
        elif opt == '-p' :
            if not os.path.isdir( arg ) :
                print "'%s' is not a valid directory." % arg
                sys.exit()
            path = arg
            if arg[-1] != '/':
                path = arg+'/'
            patternfile = path+pattern_file_name
#            combinations_file = path+'item_combinations'
	    items_file = path+'item_combinations'
    main()
