# -*- coding: utf-8 -*-
import random
from copy import deepcopy


class TreeNode:
    def __init__(self, label, tail_length, node_count=None, pattern_count=None) :
        self.nounce = random.randint(0,10000000000000000) # a "unique" identifier 
        self.label = label
        if node_count == None :
            self.count = 1
        else:
            self.count = node_count
        self.parent = None 
        self.children = []
        # Hidden labels
        # { 'query' : { {successors-1: val-1}, ... ,{successors-n : # val-n} } }
        # Successor is the next non-star nodes, or '' if there are no
        # more non-star nodes in the branch.
        self.hidden_labels = {}    
        if tail_length < 0 :
            raise Exception, "Attempting to give new TreeNode (%s) tail length < 0 (%i). Impossible." % ( self.label, tail_length )
        self.tail_length = tail_length
        if pattern_count == None :
            self.pattern_count = 0
        else :
            self.pattern_count = pattern_count


    def __str__(self):
        return "%s(z%s)(l%s)(c%i)(p%i)" % ( self.label, str(self.nounce)[:4], self.tail_length, self.count, self.pattern_count )
        #return "%s(z%s)" % ( self.label, str(self.nounce)[:4] )


    def set_child( self, child, ignore_duplicates=False, ignore_verification=False, debug=False ) :
        if not isinstance(child, TreeNode):
            raise Exception, "Child is not a TreeNode, but a %s" % type( child )
        if (ignore_duplicates == False) and (self.get_child( child.label, child.tail_length )) :
            raise Exception, "Node (%s) already has a child with label (%s) and tail lenght (%i): [%s]. \
            Can not add child %s." % (self, child.label, child.tail_length, ', '.join( self.get_children_labels() ), child )

        # If the child already has a parent, remove the child from that
        # parent's list of children
        if child.parent :
            child.parent.remove_child( child )
        child.parent = self
        self.children.append( child )

        # Make sure that the total children count does not exceed that
        # of the node itself, or that the node count is larger than the
        # total chidren count.
        if not ignore_verification :
            self.verify_nc()
            
    def get_child( self, label, tl=None ) : # Can not use tl=False, tl = 0 is valid
       
        for child in self.children :
            if child.label != label :
                continue
            if tl == None :
                return child
            if child.tail_length == tl :
                return child

    def get_children_details( self ) :
        " Returns a list with tuples (label, tail_length) "
        det = []
        for child in self.children :
            det.append("%s" % child )
        return det

    def get_children_labels( self ) :
        " Returns a list with all the node's children labels "
        lab = []
        for child in self.children :
            lab.append("%s" % child.label )
        return lab

    def increase_nc( self, value=None ) :
        if value != None :
            self.count += value
        else :
            self.count += 1
    def decrease_nc( self, value=None ) :

        if value != None :
            self.count -= value
        else :
            self.count -= 1
        if self.count < 0 :
            raise Exception, "Node %s's node count is less than 0." % self

    def increase_pc( self, value=None ) :
        if value != None :
            self.pattern_count += value
        else :
            self.pattern_count += 1
        if self.pattern_count > self.count :
            raise Exception, "%s's pattern count(%i) is higher than its node count (%i). Impossible." % (self, self.pattern_count, self.count )
    
    def decrease_pc( self, value=None ) :
        if value != None :
            self.pattern_count -= value
        else :
            self.pattern_count -= 1
        if self.pattern_count < 0 :
            raise Exception, "Node %s's pattern_count is less than 0." % self


    def calculate_successor( self, branch, layout=False, return_sub_branch=False, debug=False ) :
        " Calculates a node's non-star succeeders. "

        # TODO: A dictionary can have tuples as keys
        #       In future: use tuples instead of fooling around with strings

        if debug: print "\tCalculating successor for node %s." % self

        i = branch.index( self ) + 1

        sub_branch = branch[i:] # The branch, under node self
        n = len( sub_branch )
        successors = []

        if layout :
            if type( layout ) != type( [] ) :
                raise Exception, "Layout must be a list, it is a %s." % ( type( layout ) )
            # We're trying to find the successor as it will be after the
            # branch has been modified by the insertion of a line
            lt = layout[i:]

            for index, node in enumerate( sub_branch ) :
                if lt[index] == 0 :
                    successors.append( node.label )

        else :
            for index, node in enumerate( sub_branch ) :
                if node.label != '*' :
                    successors.append( node.label )
       
        if debug: 
            if layout:
                print "\t\tNew successors generated: ", successors
            else :
                print "\t\tSuccessors generated: ", successors
            

        if return_sub_branch :
            return sub_branch, ','.join( successors )
        return ','.join( successors )


    def rename_node( self, new_label, branch, layout, debug=False ) : # successor = list of objects
        " Renaming a node to '*', giving it hidden labels"

        old_label = self.label
        parent = self.parent
        
        if not parent :
            raise Exception, "Unless we would like to rename the root node, this is indeed an error..." 
        
        if debug:
            print "\tRenaming: Node %s will be renamed, adding hl %s." % ( self, new_label )


        # Get the branch underneath self, and corresponding successors label
        sub_branch, successors_label = self.calculate_successor( branch, layout, return_sub_branch=True, debug=debug )
        
        # First, find out whether node self has more than one
        # sub-branch. In that case, we need to split the branch.
        branches = self.dump_tree( branch_count=False )
        
        if sub_branch not in branches:
            raise Exception, "Branch is not among node %s's sub-branches." % ( self )
        
        if debug: print "\t\tNode %s has sub-branches: " % ( self ) 
        
        branch_labels = self.get_branch_labels( branches, details=True, debug=debug )
       
        # Use the pattern count of the last node in branch to
        # determine how many nodes to split out of the original branch.
        last = branch[-1]
        if not last.pattern_count :
            raise Exception, "Last node %s in branch has no pattern count." % ( last )
        
        if len( branches ) > 1 :
            if debug: print "\t\tNode %s has more than one sub-branch. Splitting required."
            self.split_branch( sub_branch, last.pattern_count )
          
            # Finished splitting out branch
            if debug: 
                print "\t\tJust finished splitting at node %s. Result is: " % self


        # First, we simply rename the node, and set the first hidden label. 
        # We look for a twin with an identical label and tail_length
        # When the second label is added, the node count is increased.
        
        if debug:
            print "\t\t Node %s is about to be renamed:" % ( self )
        
        self.label = '*'
        self.set_hidden_label( old_label, successors_label, value=self.count, increase_count=False )
        
        twin = parent.get_twin( self ) # twin, another starnode with similar tail_length

        if twin :

            if debug :
                twin.verify_nc()
                twin.verify_hl_count( debug=debug )
           
            next_index = branch.index( self ) + 1
            sub_layout = layout[next_index:]

            if 1 in sub_layout : # There are other changes following this renaming
                
                old_successors_label = self.calculate_successor( branch, debug=debug )
            
                if debug and ( old_successors_label == successors_label ) :
                    print "The new successor %s is equal the old successor %s." % ( successors_label, old_successors_label )
                    print "\tLen sub branch: ",len(sub_branch)
                    print "\tLayout: ", layout
                    print "\tSub_layout: ", sub_layout

                    raise Exception, "Successors labels are the same before calling update_successor"

                twin.update_successor( old_successors_label, successors_label, allow_absence=True, debug=debug )
            
            if debug: print "\t\tTwin of %s found: %s. It will be adopted." % (self, twin)
            # We adopt the twin, so that the references to the original objects are kept.
            self.adopt_children( twin )

            if debug: 
                print "\t\t\tAfter adoption of children, parents successors looks like this:"
            
                res2 = parent.dump_tree()
                for r2 in res2:
                    tmp2 = []
                    for s2 in r2:
                        tmp2.append("%s" % s2)
                print "\t\t\t%s" % tmp2

        if debug :
            self.verify_nc()
            self.verify_hl_count( debug=debug )

        self.set_hidden_label( new_label, successors_label )
        
        if debug: 
            print "\t%s's node count(%i)(%i) after renaming.  Hidden labels:" % ( self, self.count, self.pattern_count) , self.hidden_labels
            print "Finished renaming node %s. " % self


    def copy_node( self ) :

        new = TreeNode( self.label, self.tail_length, node_count=self.count, pattern_count=self.pattern_count )
        #new.parent = self.parent
        #new.children = self.children[:]
        new.hidden_labels = deepcopy( self.hidden_labels )
        
        return new


    def split_hls( self, to_node, branch, debug=False ) :
        " Split out hidden labels belonging to another node. "

        successor = to_node.calculate_successor( branch, debug=debug )

        # Make sure the to_node has no hidden labels before we start
        orig_hls = self.hidden_labels 
        to_hls = to_node.hidden_labels = {}

        if successor not in orig_hls :
            raise Exception, "Split hidden labels failed. Successor %s not found amongst node %s's hidden labels." % ( successor, self )
        
        if debug: 
            no_labels = len( orig_hls[ successor ] )
            print "Attempting to split out successor %s (containing %i labels), from node %s to node %s" % ( successor, no_labels, self, to_node )
        
        # TODO: Check whether deepcopy is needed in set_hidden_label and adopt_hidden_labels too!

        # Must make a copy of the dict, otherwise it will simply be a
        # reference to the same dict in orig_hls. We don't want that ...
        to_hls[ successor ] = deepcopy( orig_hls[ successor ] ) 

        # Delete the successor from the original node
        del orig_hls[ successor ]

        if debug:
            
            if no_labels != len( to_hls[ successor ] ) :
                print to_hls
                raise Exception, "Split hls failed. The successor no longer contains the same number of labels as it did before the split."

            print "\tResulting successors, self: "
            for ss in orig_hls.keys() :
                print "\t\t", ss

            print "\tResulting successors, to_node: "
            for ts in to_hls.keys() :
                print "\t\t", ts


    def split_branch( self, original_branch, val, drop_hidden_labels=False, debug=False ) :
        " Split out a branch. "

        # The total branch count is defined by the last node in the branch
        # val = original_branch[-1].pattern_count

        if type( val ) != type( 123 ) :
            raise Exception, "Branch count must be an integer, it is a %s. " % type( val )
        if type( original_branch ) != type( [] ) :
            raise Exception, "Original branch must be a list, it is a %s. " % type( original_branch )
        if val < 1 :
            raise Exception, "The number of nodes to be split must be higher than 0."

        branch = [ self ] + original_branch

        new_branch = []

        n = len( branch ) 
        
        branch_parent = None

        # We star at the bottom of the branch, making a copy of the
        # node. We decrease this node's count and attach it to its
        # original parent. We split out the original node, and modify
        # it, giving it count val, and possibly a pattern count.

        if debug :
            print "\tStarting split branch process ..."
            print "\t\tBranch: (Branch count is: %i) " % ( val )
            for b in branch :
                print "\t\t\tNode %s " % ( b )

        for index, node in enumerate( branch ) :
            
            parent = node.parent
            
            #if debug: print "Looking at node %s with parent %s" % ( node, parent )
            if debug: print "\t\tAt node %s :" % ( node )

            if node.count == val : # Safe to inherit rest of branch directly
                # It may be that the branch has only one label successor
                # connected to it, and that the entire branch should be
                # adopted. It that's the case, leave it alone.
                if branch_parent == None : 
                    if debug: print "\t\tNode %s has node count=val=%i, but has no branch parent. Leave it alone to be adopted." % ( node, node.count )
                    return
                if debug: print "\t\tNode %s has node count=val=%i. Adopting directly." % ( node, node.count )
                branch_parent.set_child( node )
                break

            # TODO: The new nodes tail length is dictated by the
            # original node, but if the remaining branch is longer than
            # the one split out, the count may become incorrect.
            new_node = node.copy_node()
            
            if debug:
                print "\t\t\tCreated new node: %s." % new_node
            
            # Correct the counts of the two nodes
            
            new_node.parent = None
            new_node.decrease_nc( val )
            node.count = val
            node.pattern_count = 0
            if index == ( n-1 ) : # we're at the last node
                if not new_node.pattern_count :
                    raise Exception, "Branch split failed. Last node (%s) does not have pattern count." % ( new_node )
                # When splitting during releasing, the branch to be
                # split can produce and identical branch, thus splitting
                # also at the very last node of the branch.
                if ( not drop_hidden_labels ) and ( new_node.pattern_count != val ) :
                    raise Exception, "Branch split failed. New node (%s) does not have pattern count %s equal to val (%s)." % ( new_node, new_node.pattern_count, val )
                new_node.decrease_pc( val )
                node.pattern_count = val
            if debug:
                print "\t\t\tCorrected count: %s" % new_node
                print "\t\t\tNode %s has new count." %  ( node )         
            
            # Children are moved from node to new_node
            
            children = node.children[:]
            
            if debug and children: 
                print "\t\t\tNode %s has children: " % node , node.get_children_details()
            
            for child in children :
                if debug: print "\t\t\t\tNew node %s adopts child: %s." % ( new_node, child )
                new_node.set_child( child, ignore_verification=True )
            
            #if debug: print "\t\tNew node %s has adopted children: " % new_node, new_node.get_children_details()


            # Split hidden labels to their respective owners :
            
            if drop_hidden_labels :
                new_node.hidden_labels = {}
                node.hidden_labels = {}
            if new_node.hidden_labels :
                new_node.split_hls( node, branch, debug=debug )

            
            # We can now verify the counts. 
            # If not at the last item, then we can not verify node
            # counts, but we can verify hidden label counts.
            new_node.verify_hl_count( debug=debug )
            node.verify_hl_count( debug=debug )
            if index == ( n-1 ) :
                new_node.verify_nc()
                node.verify_nc()
            
            # Remove the original from its parent, 
            # and replace it with the new copy.
            if index == 0 :
                #if debug: print "\t\tFirst node: Adding new node %s to parent %s." % ( new_node, parent )
                parent.set_child( new_node, ignore_verification=True, ignore_duplicates=True ) # ignore, since we've decreased its count
            else :
                #if debug: print "\t\tRemoving node %s from parent, and replacing with new node %s." % ( node, new_node )
                branch_parent.set_child( node ) # removes it from its current parent
                parent.set_child( new_node ) 

            if debug: new_branch.append( new_node )
            branch_parent = node

        if debug: 
            print "\tFinished splitting"
            print "Verifying successors in branch."
            for node in branch :
                node.verify_successors( debug=debug )
            print "Verifying successors in new branch."
            for node in new_branch:
                node.verify_successors( debug=debug )

    def adopt_children( self, cp, ignore_tail_length=False, no_starnodes=False, debug=False ) : # cp = current parent
        
        if not isinstance( cp, TreeNode ) :
            raise Exception, "Current parent must be a TreeNode, it is an instance of %s." % type(self)
        
        if ( self.label != cp.label ) :
            raise Exception, "To merge two nodes, they must have the same label.(%s vs. %s)" % ( self.label, cp.label ) 
       
        if (self.label == '*') and (not self.hidden_labels) :
            raise Exception, "%s is a starnode, but has no hiddden_labels. Impossible." % self

        if (cp.label == '*') and (not self.hidden_labels) :
            raise Exception, "%s is a starnode, but has no hiddden_labels. Impossible." % cp

        if debug: print "Adoption: Looking at self %s and cp %s." % ( self, cp )


        # Verify all the counts of cp before proceeding
        cp.verify_nc()
        cp.verify_hl_count( debug=debug )
        if no_starnodes :
            cp.verify_successors( debug=debug )

        children = cp.children[:]
        unique_labels = set( cp.get_children_labels() )
       
        for child in children :
            
            twin = self.get_twin( child, ignore_tail_length=ignore_tail_length )
            
            if twin :
                twin.adopt_children( child, ignore_tail_length=ignore_tail_length, no_starnodes=no_starnodes, debug=debug )
            else :
                if debug: print "\tAdoption: %s's child %s have no twin and will be adopted directly to %s." % ( cp, child, self )
                self.set_child( child, ignore_verification=True )
    
        for label in unique_labels :
            if label not in self.get_children_labels() :
                raise Exception, "Child with label '%s' not adopted from %s to %s!" % ( label, cp, self )

        if debug :
            self.verify_unique_children()

        
        # All potential children have been attended to
        # We now merge cp into its twin (self).

        if debug:
            print "\tAdoption: All %s's children have been moved to %s." % ( cp, self )
            print "\t\t%s is about to take over %s's properties." % ( self, cp )
            print "\t\tCurrent parent %s's has successors %s." % (cp, cp.hidden_labels.keys() )
            print "\t\t\tIt used to have children: ", unique_labels
            lbls = self.hidden_labels.keys()
            if len(lbls) < 10:
                print "\t\tBefore: %s's successors: %s " % ( self, self.hidden_labels.keys() )
            else :
                print "\t\tBefore: %s's has %i successors." % ( self, len(lbls) )
    
        
        # Self and cp have identical labels
        if cp.hidden_labels :
            self.adopt_hidden_labels( cp, increase_count=False, debug=debug )

        self.increase_nc( cp.count )
        self.increase_pc( cp.pattern_count )

        if debug: print "\t\tAfter: %s's successors %s: " % ( self, self.hidden_labels.keys() )
       
        # Verify that all counts are correct after adoption of children and properties.
        self.verify_nc()
        self.verify_hl_count( debug=debug )
        if no_starnodes:
            self.verify_successors( debug=debug )

        if debug: print "\tAdoption: %s will be removed from its parent %s, and afterwards be destroyed." % ( cp, cp.parent )
        cp.parent.remove_child( cp )
        cp.destruct()

    # TODO : obolete! Delete all references!
    def get_unique_successors( self ) :

        raise Exception, "get_unique_successors is obsolete."

        " Returns all unique label successors for a node "
        unique = []

        for label, dict in self.hidden_labels.items() :
            unique += dict.keys()

        return set( unique )
            

    def adopt_hidden_labels( self, ac, increase_count=True, debug=False ) : # ac is the child from whom the hls are to be adopted
        ' Moves all hidden labels from one node to another. '

        if debug:
            print "\t\tAdopting hls: Adopting from node %s to node %s. " % (ac, self)
            #print "\t\tAdopting hls: Before: %s's counts(%i)(%i), labels: " % (self,self.count,self.pattern_count) , self.hidden_labels
        
        l = ac.hidden_labels
        s = self.hidden_labels
        if debug: 
            ac_sum = ac.sum_hidden_labels()
            self_sum = self.sum_hidden_labels()
        
        ac_ls = ac.calculate_no_labels()
        self_ls = self.calculate_no_labels()

        new = None
        
        # Use the node with the most successors as basis
        if ac_ls < self_ls :
            if debug: print "\t\tNode %s has the biggest hl dict, using it as basis." % self
            l = self.hidden_labels
            s = ac.hidden_labels
        
        new = deepcopy( l )

        if debug:
            if max(ac_ls, self_ls) < 10 :
                print "\t\t\tNode %s has hidden labels: " % ( ac ) , ac.hidden_labels
                print "\t\t\tNode %s has hidden labels: " % ( self ), self.hidden_labels 
            else :    
                print "\t\t\tNode %s has %i hidden labels " % ( ac, ac_ls )
                print "\t\t\tNode %s has %i hidden labels " % ( self, self_ls ) 
        
        # Iterate over the shortest dict of hidden_labels.

        for successor, dict in s.items() :
            if successor not in l :
                # Copy the entire dict with labels and values
                new[ successor ] = deepcopy( dict ) # OBS! Copy
                continue
            for label, val in dict.items():
                if label not in l[ successor ] :
                    new[ successor][ label ] = val
                    continue
                # Otherwise, add the two counts
                old_val = l[ successor ][ label ]
                new[ successor ][ label ] = ( val + old_val )

                #self.set_hidden_label( hidden_label, successor, value=val, increase_count=increase_count )
        
        # Set the entire new list of hidden labels
        self.hidden_labels = new
    
        if debug:

            new_sum = self.sum_hidden_labels()
            new_ls = self.calculate_no_labels()

            if new_ls < 10 :
                print "\t\t\tAdoption result for %s: " % self, self.hidden_labels
            else:
                print "\t\t\tAdoption result for %s: (no labels) " % self, new_ls 

            if new_sum != ( ac_sum + self_sum ):
                print "Attempting to adopt hidden labels from %s to %s failed." % ( ac, self )
                raise Exception, "Adopting hidden labels failed for node %s. Sum of hls(%i) unequal to the sum of ac labels(%i) and self labels(%i)." % ( self, new_sum, ac_sum, self_sum )
        
        if increase_count :
            self.increase_nc( ac.count )

        #if debug: print "\t\tAdopting hls: After: %s's counts(%i)(%i), labels: " % (self,self.count,self.pattern_count) , self.hidden_labels



    def set_hidden_label( self, label, successor, new_successor=None, value=None, increase_count=True, debug=False ):
        
        # Successors describe the children that would normally follow a
        # hidden label had it been represented as a node. 
        # Since no branches with different tail lengths are merged, it is
        # not necessary to take tail length into account when storing
        # hidden labels.

        if type( successor ) != type( "abc" ) :
            raise Exception, "Successor must be of type string, it is a %s." % ( type( successor ) )
        if new_successor and ( type( new_successor ) != type( "abc" ) ) :
            raise Exception, "New successor must be of type string, it is a %s." % ( type( new_successor ) )

        l = self.hidden_labels
        
        if value != None:
            val = value
        else:
            val = 1

        # If a successors name has changed, all successors in hidden
        # labels with this name must be updated
        if new_successor != None: # The empty label '' actually validates to False...
            self.update_successor( successor, new_successor, debug=debug )
            successor = new_successor
        
        if successor not in l : # Successor must be created
            l[ successor ] = { label:val }

        else : # Successor already exists
            if label not in l[ successor ] :
                l[ successor ][ label ] = val # Create the new label
            else :
                l[ successor ][ label ] += val # Add to existing value
        
        if increase_count:
            self.increase_nc( val )
            self.verify_hl_count( debug=debug )



    def update_successor(self, successor, new_successor, allow_absence=False, debug=False ):

        # All successors under self's labels must be updated to reflect
        # the renaming from successor to new successor. 

        # If we attemt to update the successors in a twin that we have
        # found, the successor is not necessarily present in that node,
        # and we must allow the successor to be absent, and simply
        # return.
    
        # We update the successor simply by creating a new successor,
        # ( if it does not already exist ), giving it the hidden labels
        # from the old successors, followin the deletion of the old
        # successor. If the new successor already exist, we must move,
        # and possibly merge, hidden labels.

        if successor == new_successor :
            raise Exception, "When updating hidden labels, the successor and new successor must have different values."

        if ( not allow_absence ) and ( successor not in self.hidden_labels ) :
            raise Exception, "Successor updating failed. Successor %s is not in node %s's hidden labels." % ( successor, self )

        if successor not in self.hidden_labels :
            return

        self.verify_new_successor( successor, new_successor )

        hls = self.hidden_labels
        
        if debug:
            sum_before = self.sum_hidden_labels()
            hls_before = hls.keys()

        if debug:
            print "\t\tUpdating successors %s to %s in node %s." % ( successor, new_successor, self )
            
            print "\t\tSuccessors before rename: ", hls.keys()

        
        old_dict = deepcopy( hls[ successor ] ) # OBS! Copy

        if new_successor not in hls:
            # Create the new successor
            hls[ new_successor ] = old_dict
        else :
            # New successor already exists. Move the labels
            for label, val in hls[ successor ].items() :
                if label not in hls[ new_successor ] :
                    hls[ new_successor ][ label ] = val
                else :
                    hls[ new_successor ][ label ] += val # add to the existing value
        
        # Delete the old successor
        del hls[ successor ]

        if debug:
            sum_after = self.sum_hidden_labels()
            
            if sum_after != sum_before :
                print "Renaming from %s to %s failed." % (successor, new_successor)
                print "Successors before renaming (sum: %i) : " % ( sum_before ), hls_before
                print "Successors after renaming (sum: %i) : " % ( sum_after ), hls.keys()
                raise Exception, "Renaming of successors failed. The sum of hidden labels has changed from %i to %i." % ( sum_before, sum_after )
            
            print "\t\tSuccessors after rename: ", hls.keys()


    def remove_child( self, child ) :
        if not isinstance(child, TreeNode) :
            raise Exception, "Child to remove should be a TreeNode, not %s" % type(child)
        if child not in self.children :
            raise Exception, "Child to be removed, %s with tail_length %i, does not exist among %s's children" % (child, child.tail_length, self )
        #print "Node %s is now being removed from its parent %s" % (child,self)
        self.children.remove( child )

    def destruct( self ) :
        if self in self.parent.children :
            raise Exception, "Forgot to detatch me (%s) from parent (%s) before deletion" % (self, self.parent )
        del self

    def sum_children( self ) :
        total = 0
        for child in self.children :
            total += child.count
        return total

    def sum_hidden_labels( self ):
        total = 0 
        for dict in self.hidden_labels.values() :
            total += sum( dict.values() )
        return total
    

    # TODO: remove all references to calculate_no_labels
    def sum_no_successors( self ) :  
        " Counts the number of labels belonging to a node's successors. "
       
        raise Exception, "sum_no_successors method is depricated!"

        total = 0
        for label, dict in self.hidden_labels.items() :
            total += len(dict)

        return total
    
    def calculate_no_labels( self ) :  
        " Counts the number of labels belonging to a node's successors. "

        total = 0
        for dict in self.hidden_labels.values() :
            total += len(dict)

        return total



    def remove_stars( self, debug=False ) :
        " Removes starnodes from a node's subrees "
        
        # Must make copies of children-list, since it may change during
        # execution.

        #self.verify_successors( "Before removing starnodes, at node %s :" % self )

        children = self.children[:]
        parent = self.parent
        
        for child in children :
           child.remove_stars() 

        if self.label != '*' :
            return

        children = self.children[:]
        
        # For each child we adopt into the parent, the count of the
        # starnode is decreased in accordance with the child's count.
        # This way, the parent's count can be verified along the way.
        for child in children :
            if debug: print "\tNode %s adopts child %s" % (parent, child)
            self.decrease_nc( child.count )
            parent.set_child( child, ignore_duplicates=True, debug=debug )
        
        # Parent should adopt all hidden labels from starnode, as long
        # as it is not a starnode itself (will cause problems with
        # counts etc.)
        # The parent's count should not be increased.
        # If self is a starnode with pattern_count, the pattern_count
        # must be moved upwards to the parent.
        if parent.label != '*' :
            parent.adopt_hidden_labels( self, increase_count=False, debug=debug ) 
        if self.pattern_count :
            parent.increase_pc( self.pattern_count )

        # All children and hidden labels are now adopted out and we can
        # gracefully perform selfdestruction
        parent.remove_child( self )
        self.destruct()


    def merge( self, debug=False ) :
      
        #children = self.children[:]
        unique_labels = set( self.get_children_labels() )

        # When merging, all children with similar labels will be
        # merged. To make sure that the iteration is correct, we iterate
        # only over one of the children with a given label, and merge
        # the other children with similar labels into this.

        #for child in children :
        for label in unique_labels :
            child = self.get_child( label ) # simply returns the first child with this label
            child.merge( debug=debug )

        parent = self.parent

        if not parent :
            return # We're at the root node
       
        # We no longer care about node tail lengths
        twin = parent.get_twin ( self, ignore_tail_length=True )
        
        self.verify_successors( "Before potential merge of node %s :" % self, debug=debug )
        
        while twin :
            if debug: print "Merging node %s" % self
            #twin.adopt_children( self, ignore_tail_length=True, debug=debug )
            self.adopt_children( twin, ignore_tail_length=True, no_starnodes=True, debug=debug )
            
            self.verify_successors( "After merge of node %s with %s :" % (self,twin), debug=debug )
            
            twin = parent.get_twin( self, ignore_tail_length=True )
        

    def release_hls( self, hl_release_limit, debug=False ) :
        " If hidden labels in occur amongst the node's children labels, they should be released to the existing children."
  
        #print "Node %s has children : " % self , self.get_children_labels()
        #print "And successors: ",self.hidden_labels.keys()
        #self.verify_successors()
        #print "\tSuccessors verified."

        hls = self.hidden_labels
        
        if hls :
            successors = hls.keys()

            for successor in successors :
                dict = hls[ successor ]
                children = self.children
                if not self.children :
                    continue

                visited = []
                child = children[0] # Simply fetch the first child
            
                  
                while ( dict and child ) :

                    visited.append( child )

                    # There are no children with identical labels, thus,
                    # if a successor's label is the same as the child
                    # label, then there can be no node to adopt!
                    if debug:
                        if child.label == successor :
                            print "\tSuccessor and child label is the same (%s). Continuing without doing anything." % successor
                    

                    if ( child.label != successor ) and ( child.label in dict ) :
                        
                        val = dict[ child.label ]
                        total = sum( dict.values() )
                    
                        if val < ( total * hl_release_limit ) :
                            remaining = list( set( self.children ) - set( visited ) )
                            if not remaining :
                                break # We have been through all the labels
                            child = remaining[0]
                            continue
                            
                        if debug: 
                            print "\tLooking at child %s, successor %s" % ( child, successor )
                            print "\t\tBefore releasing label %s, node %s has children: " % (child.label, self), self.get_children_details()
                        child.receive_hl( successor, debug=debug )
                    
                        # If we are at the last label, we break it off
                        if successor not in hls :
                            dict = {}
                            if debug: print "\t\tAt the last label in successor %s. All hidden labels released. " % successor
                            break 

                        # If the child label turned out to be the next
                        # to last label in the successor, we also
                        # release the remaining label.

                        # The idea here is that it is hard to decide
                        # whether a label is a keyword or a variable. If
                        # one label is present among one of the node's
                        # children labels, then it is very likely that
                        # the other label is also a keyword.
                        if len( hls[ successor ] ) == 1 :
                            if debug: print "\tLast label discovered."
                            self.handle_last_hl( successor, debug=debug )

                        if debug: print "\t\tAfter releasing label %s, node %s has children: " % (child.label, self), self.get_children_details()

                        dict = hls[ successor ]
                    
                    remaining = list( set( self.children ) - set( visited ) )
                    if not remaining :
                        break # We have been through all the labels
                    child = remaining[0]
        

        # Now all hidden labels are handled (or there simply were none), 
        # we can iterate over the remaining children
       
        #print "Verfify that all remaining successors are correct for node %s, before handling the kids." % self
        #self.verify_successors()
        
        for child in self.children :
            child.release_hls( hl_release_limit, debug=debug )
        
        #print "Verfify that all remaining successors are correct for node %s after handling the kids." % self
        #self.verify_successors()

    def receive_hl( self, successor, debug=False ) :
        " Release a parent's hidden label to the node. "

        # There are two types of hls:
        # label:{'':val}            # No more non-star nodes to be found in branch
        # label:{'A,B,C,D':val}
    
        parent = self.parent
        parent_labels = parent.hidden_labels[ successor ]
        val = parent_labels[ self.label ]
        
        if debug :
            self.verify_nc()
            self.verify_hl_count( debug=debug )
            self.verify_successors( debug=debug )
        
            print "\tReceive hl: Node %s, successor %s and value %i: " % ( self, successor, val )

        # Simplest case
        if successor == '' :
            # If successor is '', it means that the starnode it once
            # belonged to had the pattern count set, which was again
            # inherited to the current holder. Thus, the pc must be
            # decreased when such a label is removed.
            self.increase_nc( val )
            self.increase_pc( val )
            parent.decrease_pc( val )
            parent.remove_successor_label( successor, self.label )

            parent.verify_hl_count( debug=debug )
            parent.verify_pc()
            parent.verify_nc()

            return


        # Complex case 

        if debug:
            print "\t\tParent node %s has the following branches: "  % parent
        parent_branches = parent.dump_tree( branch_count=False, debug=debug ) # [ [node objects], ..., [node objects] ]
        parent_branch_labels = parent.get_branch_labels( parent_branches, debug=False ) # preserves order

        index = parent_branch_labels.index( successor )
        successors = parent_branches[ index ]  # We find the branch

        ac = successors.pop(0) # Adoptive child, the first child in the successor branch
        if debug:
            print "About to split at node %s and branch: " % ac, successors
        # When splitting a branch at this point, we drop the hidden
        # labels in the nodes we are about to split. This is because
        # there is no way of telling to which of the branches the hidden
        # labels belong to.
        if debug:

            print "\t\tBefore split at node %s, with val %i, parent %s has the following branches:" % ( ac, val, parent )
            before_branches = parent.dump_tree( branch_count=False, debug=debug )
            before_branch_labels = parent.get_branch_labels( before_branches, details=True, debug=debug ) # preserves order
        
        ac.split_branch( successors, val, drop_hidden_labels=True, debug=debug )

        if debug:

            print "\t\tAfter split at node %s, parent %s has the following branches:" % ( ac, parent )
            after_branches = parent.dump_tree( branch_count=False, debug=debug )
            after_branch_labels = parent.get_branch_labels( after_branches, details=True, debug=debug ) # preserves order

        # Increase the count of the node
        # ( Must do this after getting parent's branches, otherwise the
        # dump-tree verification will fail.
        self.increase_nc( val )
        if debug:
            print "Node %s had its node count increased." % self

        # Find out if self has a child with identical label as ac
        twin = self.get_twin( ac, ignore_tail_length=True )
               
        if twin:
        
            if debug: print "\t\tTwin node %s will adopt %s." % ( twin, ac )
            twin.verify_nc()
            if debug: print "\t\t\tTwin %s before merge." % ( twin )
            twin.adopt_children( ac, ignore_tail_length=True ) # OBS!  Remember to ignore tail length!
            if debug: print "\t\t\tTwin %s after." % ( twin )
            ac = twin

        else :
            self.set_child( ac )
            if debug:
                p = ac.parent
                print "\t\tNode %s has no other children like %s, adopted it directly." % ( self, ac )
                print "\t\tSelf %s has now got children: " % self, self.get_children_details()

        if debug:
            print "\t\tAfter adoption, %s has the following properties: " % ( ac )
            print "\t\t\tParent: %s" % ac.parent
            print "\t\t\tChildren: ", ac.get_children_details()

            print "\t\tVerifying the properties of old parent %s: " % parent
        parent.verify_nc()
        parent.verify_hl_count( debug=debug )

        parent.remove_successor_label( successor, self.label )
        
        if debug: 
            print "Tree now looks like: "
            after_branches = parent.dump_tree( branch_count=False )
            after_branch_labels = self.get_branch_labels( after_branches, details=True, debug=True )


    def handle_last_hl( self, successor, debug=False ) :
        
        # Now that we have removed a successor, it would be
        # interesting to see if there exists only ONE last hidden
        # label with the same successors, whose label does not occur
        # amongst the children of self.
        # If only one such label exists, it should be released to
        # adopt the (correct number of nodes from the) branch.

        if debug: print "Looking for a last label in succssor: ", successor

        if successor not in self.hidden_labels :
            if debug: print "\tSuccessor not found in hidden labels."
            return # No more labels in successor, successor has been removed

        successor_labels = self.hidden_labels[ successor ]    
            
        if len( successor_labels ) > 1 :
            if debug: "\tLength of successor labels > 1: ", successor_labels
            return

        new_label = successor_labels.keys()[0] # get the name of the remaining label
        
        if new_label in self.get_children_labels() :
            # It will be handled by the initial iteration
            if debug: print "\tThe new label %s is already present among %s's children." % ( new_label, self)
            return

        if debug:
            print "\t\tNew node with label %s will be added to node %s" % ( new_label, self )

        # Node count is set to 0. It will be increased when
        # the label is released from the parent. 
        # A symbolic tail length is set.
        tl = ( self.tail_length - 1 )  # Even though it is not needed anymore
        new_node = TreeNode(new_label, tl, node_count=0 )
        self.set_child( new_node ) # will set parent to self in new_node
        
        if debug: 
            print "Tree now looks like: "
            branches = self.dump_tree( branch_count=False )
            branch_labels = self.get_branch_labels( branches, details=True, debug=True )
        
        #new_node.receive_hl( successor )
            

    def remove_successor_label( self, successor, label ) :
        hls = self.hidden_labels

        if successor not in hls :
            raise Exception, "Deletion attempt failed. Successor %s not present among node %s's hidden labels. " % ( successor, self )
        if label not in hls[ successor ]:
            raise Exception, "Deletion attempt failed. Label %s not present among successor %s's labels in node %s. " % ( label, successor, self )
        
        del hls[ successor ][ label ]
        # It the last label in successor has been deleted, delete the successor as well
        if not hls[ successor ]:
            del hls[ successor ]


    # TODO: remove depricated 
    def remove_label_successors( self, label, successors ):
       
        raise Exceptio, "Method remove_label_successors is depricated."

        hls = self.hidden_labels
        if label not in hls :
            raise Exception, "Label %s is not present among node %s's hidden labels. Attempt to delete it failed." % (label, self)
        if successors not in hls[label]:
            raise Exception, "Successor %s is not present among label %s's successors on node %s." % (successors, label, self)
        del hls[label][successors]
        # If label has no more successors, delete the entire label from self's hidden labels
        if not hls[label] : 
            del hls[label]

    def get_twin( self, ac, ignore_tail_length=False ) : # ac = adoptive child
        " Looks for a node with similar label (and possibly tail length) as node to be adopted, but not the child itself "

        children = self.children

        if children.count( ac.label ) == 1:
            # child is the only node with this label
            return
        for child in children :
            if ( child.label == ac.label ) and ( child != ac ) :
                if ignore_tail_length :
                    # Return the first occurrence of a match
                    return child
                elif ( child.tail_length == ac.tail_length ) :
                    return child

    def dump_tree( self, branch_count=True, find_matching=False, depth=0, debug=False ):
        " Returns every branch found beneath node"
        
        # Must include depth to make sure that, if the node that first
        # calls this method has pattern count, it will not return an
        # empty list as the first branch.
        # Also need it to return branches of correct length when finding
        # matching branches.

        # We are never interested in branch counts when we are looking
        # to find a good match amont the branches. (Precaution).
        if find_matching :
            branch_count = False
         
        children = self.children
       
        if ( len(children) == 0 ) and ( self.pattern_count != self.count )  :
            raise Exception, "A node(%s) without children must have pattern count(%i) equal to node count(%i)." % ( self, self.pattern_count, self.count )
        
        #print "\tDepth is %i and node %s has children:" % (depth,self) , self.get_children_labels()

        res = []

        # If find_matching is set, we wish to find only branches of the
        # same length as the line, thus we simply return if we find a
        # branch that is too short, or too long.
        if find_matching :
            if ( depth != 0 ) and self.pattern_count :
                # We are not yet as deep as we like, but the pattern ends here
                return []
            if ( depth == 0 ) and not self.pattern_count :
                # We have come as deep as we would like, but pattern has not ended
                return []
            if ( depth == 0 ) and ( len( children) == 0 ) :
                return [ [] ]
            if ( depth == 0 ) and self.pattern_count :
                res.append( [] )

        else :
            # Needed since a node may have pattern count set, and still have
            # children.
            if len( children ) == 0 :
                if branch_count :
                    return [ [str( self.count )] ]
                return [ [] ]
        
            # When listing all children underneath a node,  must check when depth == 0, 
            # otherwise, if the node we're first calling the method on has pattern
            # count set, it will return an empty list as the first result.
            # Not good. We ignore this test if find_matching is set.
            if ( depth > 0 ) and ( self.pattern_count > 0 ) :
                if branch_count :
                    res.append( [str( self.pattern_count )] )
                else :
                    res.append( [] )
        
        for child in children:
            if find_matching :
                sub_branches = child.dump_tree( find_matching=find_matching, depth=(depth-1) )  # list containing one or more lists
            else:
                sub_branches = child.dump_tree( branch_count=branch_count, depth=(depth+1) )  # list containing one or more lists
            for grandchildren in sub_branches:
                res2 = [ child ] + grandchildren
                res.append( res2 )
        return res


    def get_branch_labels( self, branches, details=False, debug=False ) : 
        " Takes a list of branches and returns a string representation of the nodes in the branches."

        if type(branches) != type([]) :
            raise Exception, "Can not get branch labels, branches must be a list, it is a: " % type(branches)

        branch_labels = []

        for branch in branches:
            b = []
            for node in branch :
                if not details :
                    b.append( node.label )
                else :
                    b.append( "%s" % node )
            if not details :
                branch_labels.append( ','.join( b ) )
            else :
                branch_labels.append( b )

        if debug :
            for l in branch_labels :
                print "\t\t\t",l
        
        return branch_labels    


    # Code to verify that the calculations along the way are always correct.

    def total_verification( self, verify_unique_children=False, verify_successors=False ) :
        
        for child in self.children :
            child.total_verification()

        self.verify_nc()
        self.verify_pc()
        self.verify_hl_count( debug=True )
        if verify_successors:
            self.verify_successors( debug=True )
        if verify_unique_children :
            self.verify_unique_children()


    def verify_successors( self, text=None, debug=False ) :
        " Verifies that all the node's successors actually exists underneath it."

        # Verification of hidden labels is very expensive when the
        # number of hidden labels grows large. Only perform during debug
        # mode.
        if not debug:
            return

        # Verification of successors can not be performed until stars
        # have been removed!

        hls = self.hidden_labels

        if not hls :
            return

        branches = self.dump_tree( branch_count=False ) # [ [node objects], ..., [node objects] ]
        branch_labels = self.get_branch_labels( branches ) 

        successors = hls.keys()
        
        if '' in successors :
            # Calculate the sum of the labels under this successor, and
            # make sure the node's pattern count >= this sum.
            tot = sum( hls[''].values() )
            if tot > self.pattern_count :
                raise Exception, "Successor verification failed. Number of \'\'-successors (%i) is larger than the node %s's pattern count (%i)." % (tot, self, self.pattern_count )
            successors.remove('') # since the successor '' will never be found as a branch

        for successor in successors :
            if successor not in branch_labels :
                print "\t\t\t\t\tWE HAVE A PROBLEM!"
                if debug:
                    if text :
                        print text
                    print self.breadcrumbs2()
                    print "\tSuccessor '%s' not found: Node %s has sub-branches: " % ( successor, self)
                    dump = self.dump_tree( branch_count=False )
                    self.get_branch_labels( dump, debug=True )
                    print "%s's parent is node %s." % (self,self.parent)
                    print "Parent %s has got hidden labels: " % self.parent, self.parent.hidden_labels
                    

                    for b in branch_labels:
                        print "\t\t%s" % b
                raise Exception, "Successor %s not found amongst node %s's sub-branches!" % ( successor, self )

    # Should not be part of the total verification.
    def verify_new_successor( self, successor, new_successor ) :
        " Verifies that the new successor is a subset of the previous successor. "
        
        if new_successor == '':
            return

        os = successor.split(',')
        ns = new_successor.split(',')

        for item in ns :
            if item not in os :
                # new_successor is not a subset of successor
                raise Exception, "New successor(%s) is not a subset of successor(%s)." % ( new_successor, successor )


    def verify_nc( self ) :

        sum = self.sum_children()

        s = []
        for child in self.children:
            s.append( "%s(%i)" % (child.label, child.count) )

        if sum != ( self.count - self.pattern_count ) :
            print "\tnc error: Node %s has pattern_count: %i:" % (self, self.pattern_count)
            print "\tnc error: Node %s's children and their counts: " % self, s
            raise Exception, "Sum of children counts (%i) is not equal to node %s's count - pattern count (%i - %i). " % (sum, self, self.count, self.pattern_count)
    
    def verify_pc( self ) :
            
        if ( not self.children ) and ( not self.pattern_count ) :
            raise Exception, "%s has no children, but has no pattern count."
        
        if not self.pattern_count :
            return
        sum = self.sum_children()

        if self.pattern_count != ( self.count - sum ) :
            raise Exception, "%s's pattern count(%i) is unequal to the node's count(%i) - the sum of the children's counts(%i). " % ( self, self.pattern_count, self.count, sum )


    def verify_hl_count( self, debug=False ) :

        # Verification of hidden labels is very expensive when the
        # number of hidden labels grows large. Only perform test in
        # debug mode. The total verification will identify problems
        # after each stage.
        if not debug :
            return

        # We only control starnodes, since a normal node who has
        # inherited hidden labels from one or more starnodes may have a
        # hidden label count different from its node count.
        if self.label != '*':
            return
     
        sum = self.sum_hidden_labels()
        if sum != ( self.count ) :
            raise Exception, "Hidden label verification failed for node %s.\nSum of hidden labels(%i) is unequal to node count(%i). " % (self, sum, self.count ) 

    def verify_merge( self ) :
        # Naive, but works
        for child in self.children :
            child.verify_merge()
        self.verify_unique_children()

    def verify_unique_children( self ) :
        unique = []
        for child in self.children :
            if child.label not in unique :
                unique.append(child.label)
            else :
                print "Node %s has ducplicate children." % self
                print "Hidden labels: ", self.hidden_labels
                print "Children: ", self.get_children_details()
                raise Exception, "Node %s has duplicate children labels: %s." % ( self,self.get_children_labels() )



    # Some code for aiding the debug process...

    def get_children_str( self ) :
        l = []
        for child in self.children:
            l.append( "%s" % child )
        return l

    def breadcrumbs2( self ) :
        return self.breadcrumbs() +"\t\t--->\t"+ str(self.get_children_str())

    def breadcrumbs( self ) :
        "Print path from root"
        if not self.parent:
            return "ROOT"
        r = self.parent.breadcrumbs()
        #return "%s > %s" % (r, self.label)
        #l = self.label
        l = self
        #if self.label == '*':
        #    l = self
        return "%s > %s" % (r, l)
