# -*- coding: utf-8 -*-

from msg_extraction import extract_message
from difflib import get_close_matches
from TreeNode2 import TreeNode
from tree_graph import draw_tree, draw_tree2
import sys, os.path, pickle, re, getopt, ConfigParser


def usage() :
    print "\nUsage of patter mining-based log analyser: "
    print "\t-h\t(--help) To show this message."
    print "\t-l\tFull path to the logfile to analyse (mandatory)."
    print "\t-p\tFull path to the directory where the results should be written (mandatory)."
    print "\t-d\t(--debug) To turn on debugging code."
    print "\t-c\t(--config) Full path to a different config file, the default is 'TreeStructure.config'."
    print "\t-s\t(--stages) Print out results from all stages during analysis."
    print "\t-r\t(--progress) Print progress data."



def main( ) :    

    global logfile, patternfile, graph_node_threshold, exceptions_file
    global debug, stages, progress

    # When testing the code, we don't want to read the entire log every
    # time we make a change, thus we create a cache
    bname = os.path.basename(logfile)
    tmpname = "/tmp/%s.pickle" % bname
    
    
    print "Starting ..."

    root = TreeNode( "LogTree", 0, node_count=0 ) # Faking tail_lengt to 0
 
    # First register all known spam patterns
    if exceptions_file :
        spampatterns = []

        s = open( exceptions_file, 'r' )
        for exp_line in s:
            spampatterns.append( exp_line )
        s.close()

 
    # Keep track of the number of lines read, and how many of these
    # lines were considered spam and thus ignored.
    # TODO: Spam-patterns should be moved out to separate file!
    spam_count = 0
    line_count = 0

    if os.path.exists(tmpname) and (os.stat(tmpname)[8] > os.stat(logfile)[8]):
        root = pickle.load(open(tmpname))
        print "Using cached tree structure from file %s" % tmpname
    else:
        print "Cache file doesn't exist or is older than logfile. Reparsing log .."
        f = open(logfile, 'r')


        for line in f:

            #if (line_count) - spam_count > 99999:
            #    break
            
            if len( line.strip() ) == 0:
                # Line is empty, jump gracefully over it.
                continue

            line_count += 1
            
            # Simply to keep track of progression 
            if progress :
                if (line_count % 1000) == 0 :
                    print "Passed %d lines" % line_count
            
            # Prune away spam 
            # MDLOG contains a lot of spam-subject (usually in combination
            # with, spam, accepted, passed, or grey. )
            if exceptions_file :
                spam = False
                for spampattern in spampatterns :
                    sp = re.compile( spampattern )
                    if sp.match( line ) :
                        spam = True
                        spam_count += 1
                        break
                if spam :
                    continue

            msg = extract_message( line )

            node, line, tail_length, branch, layout = find_matching_branch( root, msg, debug=debug )
          
            if node == None : # Nothing to insert
                continue

            add_to_tree( node, line, depth=tail_length, branch=branch, layout=layout, debug=debug )


        f.close()
        pickle.dump(root, open(tmpname, 'w+'))

    print "\nFinished inserting loglines from log file '%s' into tree structure ... " % bname
    print "\tNumber of lines inserted into tree: %i ( total %i lines read)" % ( line_count-spam_count, line_count )
    print "\tNumber of spam-messages found: %i" % spam_count
    
    # Draw graph and write results to file
    # Rename to png file type if wanted

    n1 = '_1_star_patterns'
    n2 = '_2_non_star_patterns'
    n3 = '_3_merged_patterns'
    f1 = patternfile + n1
    f2 = patternfile + n2 
    f3 = patternfile + n3
    f4 = patternfile
    g1 = patternfile + n1 + '.svg'
    g2 = patternfile + n2 + '.svg'
    g3 = patternfile + n3 + '.svg'
    g4 = patternfile + '.svg'
   

    if stages :
        # Star-patterns
        print "Saving starpatterns ..."
        save_patterns( root, f1 )
    
        try :
            os.unlink(g1)
        except :
            pass
        draw_tree( root, g1, graph_node_threshold )

    # Removing stars
    print "Pruning away starnodes ..."
    root.remove_stars( debug=debug )

    if stages :
        print "Saving non-star patterns ..."
        save_patterns( root, f2 )
    
        try :
            os.unlink(g2)
        except :
            pass
        draw_tree( root, g2, graph_node_threshold )
    
    # Verification of tree ( after removal of stars )
    if debug: print "Performing total verification, after removal of stars."
    root.total_verification()
   
    # Merging
    print " Merging similar branches... "
    root.merge( debug=debug )

    if stages :
        print "Saving merged patterns ..."
        save_patterns( root, f3 )
    
        try :
            os.unlink(g3)
        except :
            pass
        draw_tree( root, g3, graph_node_threshold )
    
    # Verification of tree ( after merge )
    if debug: print "Performing total verification, after merge."
    root.total_verification( verify_unique_children=True, verify_successors=True )
    
    print " Releasing hidden labels... "
    global hl_release_limit # from configuration file
    root.release_hls( hl_release_limit, debug=debug )

    print "Saving patterns ..."
    save_patterns( root, f4)
    
    try :
        os.unlink(g4)
    except :
        pass
    draw_tree( root, g4, graph_node_threshold )
   
    # We cannot verify a node's hidden label successors after releasing
    # hidden labels, since the paths may changed. ( As an example,
    # pattern counts may decrease to zero ).
    if debug: print "Performing total verification, after releasing."
    root.total_verification( verify_unique_children=True )

    print "Finished!"


    


def save_patterns( root, file ):

    f = open(file, 'w+')

    patterns = root.dump_tree()
    #patterns = map_out_tree(root)

    for pattern in patterns:
        #line = ' '.join(pattern)
        count = pattern.pop()
        l = []
        for node in pattern :
            l.append( node.label )
        l.append( count )
        line = ' '.join( l )
        #line = ' '.join( patterns )
        f.writelines(line+'\n')

    f.close()


def find_matching_branch( root, line, debug=False ) :

    stem = []
    stem_labels = []
    sub_branches = []
    node = root

    maxlen = len( line )
    if debug: print "Inserting line: ", line

    # For each node we attach to root, we must increase its count.
    if node == root :
        node.increase_nc()

    for index, item in enumerate( line ):

        tail_length = ( (maxlen - 1) - index )
        depth = ( maxlen - index )
        child = node.get_child( item, tail_length )

        if child :
            stem.append( child )
            stem_labels.append( child.label )
            # We increase the count of the nodes in the stem immediately
            child.increase_nc()
            node = child
            if index == ( maxlen - 1 ) : # at the last item in line
                child.increase_pc()
                return ( None, None, None, None, None )
            continue

        # If we do not find the first item in the list, we will insert the
        # entire line directly.
        if node == root :
            if debug: print "\tFirst line item not found. Line inserted into tree directly: ", line
            return ( root, line, None, None, None )

        # We found no matching node, so we take a look in the last node
        # in stem to see if we can find a branch to merge with.
        if debug: print "\tFind sub-brances under node %s, with depth %i...." % ( node, depth )
        sub_branches = node.dump_tree( node, find_matching=True, depth=depth )

        break
   
    stl = len( stem )
    subline = line[stl:] # The remains of line that is not part of stem
   
    # This is probably not necessary anymore, since tail length was
    # introduced.
    if ( len(sub_branches) == 0 ) :
        # If no sub_branches are found, we found no branches of length
        # maxlen, and we therefore need to insert the subline directly
        # under the last matching node. Add tail_length to know how many
        # nodes are in the stem.
        if debug: print "\tNo branch of correct length found in tree for line: ", stem_labels, '->', subline
        return ( node, subline, tail_length, None, None )

    branch, layout = find_best_branch( line, subline, stem, stem_labels, sub_branches, debug=debug)
    
    if branch or ( len( stem ) < 2 ) :
        return ( node, subline, tail_length, branch, layout ) 

    # There were no good matches under the node 
    # Perhaps node represents a variable?
    # Look under the previous stem node and see if a good match can be
    # found there.

    stem2 = stem[:-1]
    stl2 = len( stem2 )
    subline2 = line[stl2:]
    stem_labels2 = stem_labels[:-1]
    node2 = stem2[-1]
    tail_length2 = tail_length + 1
    depth2 = depth + 1 # Need to move one step higher in the tree
    
    if debug:
        print "\tSearching for a better match among %s's parent %s. " % ( node, node2 )
        print "\tShorter stem is: ", stem_labels2
        print "\tTail length is %i and new depth is %i " % ( tail_length2, depth2 )

    sub_branches2 = node2.dump_tree( node2, find_matching=True, depth=depth2 )

    branch2, layout2 = find_best_branch( line, subline2, stem2, stem_labels2, sub_branches2, debug=debug )

    if not branch2 :
        # Return the original
        if debug: print "\tInserting subline in its entirety under node %s." % node 
        return ( node, subline, tail_length, branch, layout )

    if debug: 
        print "\t\tNode %s had no good sub-branches. Will insert under the previously found stem node %s." % ( node, node2 )
        print "\t\tDecreasing the current node %s's node count." % node
    node.decrease_nc()  # Must decrease the previously increased count.
    return ( node2, subline2, tail_length2, branch2, layout2 )
    



def find_best_branch( line, subline, stem, stem_labels, sub_branches, debug=False ) :

    stl = len( stem )
    sll = len( subline )
    node = stem[-1]
    tail_lenght = ( node.tail_length - 1 ) # Tail length is always one shorter than the parent's

    branches_keys = []
    branches_l = {} 
    for branch in sub_branches :
        tmp = []
        for n in branch :
            tmp.append( n.label )
        branches_keys.append(tmp)
        label = ' '.join( tmp )
        branches_l[label] = branch

    # Find closest method attempts to find the branches that are the
    # best fit for the line.
    
    if debug: print "\tBranches found: ", branches_keys
    candidates = find_closest( line, subline, stem_labels, branches_keys, sll, stl )
    
    max_common_fraction = 0
    best_branch = []
    best_branch_labels = None
    layout = []

    if debug: print "\tFinding candidates for line: ", stem_labels, '->', subline
    for candidate in candidates:

        if debug: print "\t\tCandidate: " , candidate
        c_label =  ' '.join(candidate)
        branch = branches_l[ c_label ]
        common_fraction, tmp_layout = calculate_likeness( subline, branch, sll, stl )

        if common_fraction > max_common_fraction:
            max_common_fraction = common_fraction
            layout = tmp_layout
            best_branch = branch
            best_branch_labels = c_label

    if max_common_fraction == 0:
        # No suitable branches found, we must insert line directly under node.  
        if debug: print "\tNone of the candidates under node %s were good enough." % node
        return ( None, None )
   
   
    if debug: print "\tBest branch found: [ %s ], with layout:" % best_branch_labels, layout 
    return ( best_branch, layout )




def calculate_likeness( subline, branch, sll, stl ) :

    # stl = lenght of stem
    # sll = length of subline

    if ( sll != len(branch) ) :
        raise Exception, "Something is wrong. Branch length(%i) is not eqal to subline length(%i)." % ( len(branch), sll )
    
    line_length = ( stl + sll )
    common = []

    # This is a prototype. To allow the unittests to call this method
    # directly, we allow setting the default values whenever the
    # global setup variable is not set.

    try :
        global setup
        if setup:
            pass

        global short_line      
        global very_short_line
        global short_subline
        global very_short_subline
    
        global min_common_fraction
        global long_stem_fraction
        global short_stem_common_fraction
        global long_stem_common_fraction
        global min_matches_required
        global max_changes_allowed
        global short_min_matches_required
        global short_max_changes_allowed
    except : # Set default values
        short_line = 8   
        very_short_line = 3
        short_subline = 3
        very_short_subline = 2
    
        min_common_fraction = 0.6
        long_stem_fraction = 0.66
        short_stem_common_fraction = 0.75
        long_stem_common_fraction = 0.75
        min_matches_required = 0.4
        max_changes_allowed = 0.5 
        short_min_matches_required = 0.3
        short_max_changes_allowed = 0.4
        
    # First we find out how many of the items in the line and branch
    # correspond to each other.
    # We do not count stars as a mismatch, but we don't want to count
    # them as matches either, since this will 'encourage' the generation
    # of starnodes. We thus treat them as a separate group.
    for i in range( sll ):
        if subline[i] == branch[i].label:
            common.append(0)
        elif branch[i].label == '*':
            common.append(2)
        else:
            common.append(1)

    # Matches are the total number of nodes that will stay unchanged in
    # the branch, were we to merge the subline with it. Not including
    # stars.
    matches = stl + common.count(0) 

    # Mismatches are the number of nodes that will have to change. Not
    # including stars. We would like this to be as small as possible.
    mismatches = common.count(1)

    # Significants include all but the stars. They are defined to be
    # insignificant, as we care little about them.
    significants = matches + mismatches
    
    stars = common.count(2)

    # To find a measure for how much the line and the branch resembles
    # each other, we look at how many matches there are per
    # matches+mismatches.
    common_fraction = ( float(matches) )/significants

    
    # If this common fraction is less than the minimum we have set, then
    # declare that this is not a good match, and return a common
    # fraction of 0. Returning 0 and not the actual fraction assures us
    # that the branch will not be considered a candidate.

    if common_fraction < min_common_fraction:
        return 0, common
    
    if ( float( stl )/line_length > long_stem_fraction ) and ( common_fraction >= long_stem_common_fraction ) :
        return common_fraction, common

    # If the length of the stem matches or exeeds the length of the
    # subline, there's a chance the stem will determine whether the
    # line and branch are similar or not. 
    # We must therefore investigate the similarity between the subline
    # and the sub-branch alone.

    base = float( sll - stars ) # All values in subline that are not stars

    sub_matches = ( matches - stl ) # Number of matches for subline
    
    # A subline may be long or short. In very short sublines, it may be
    # necessary to allow more stars or mismatches than usual. 
    matches_limit = min_matches_required

    if sll <= short_subline :
        matches_limit = short_min_matches_required
    
    matches_fraction = ( float( sub_matches )/sll)

    change_limit = max_changes_allowed  # Number of allowed changes

    # If the total line lenght is less than eight(default val) (rather short), then
    # allow a lower rate of changes to occur.
    if ( line_length < short_line ): 
        change_limit = short_max_changes_allowed

    change_fraction = float( mismatches + stars )/( line_length )

    # If line is very short, ( it has already prooved to be > 0.6
    # similar ) we declare it a good fit.
    if line_length <= very_short_line :
        return common_fraction, common

    # If similarity between line and branch is > 75% and subline is
    # short, we declare it a good match. Note that the stem in this case
    # is not considered a long stem. (As it was handled earlier).
    if ( common_fraction > short_stem_common_fraction ) and ( sll <= very_short_subline ):
        return common_fraction, common

    # If the number of changes needed are too high, we declare the branch to be
    # insufficcien.
    if (change_fraction > change_limit):
        return 0, common

    # If the the number of clear matches are too low, we declar the
    # branch to be insufficient.
    if matches_fraction < matches_limit :
        return 0, common

    return common_fraction, common



def find_closest( line, subline, stem, branches, sll, stl ) :

    candidates = []

    if type(stem) != type([]):
        raise Exception, "Stem is not a list as required, but a %s." % ( type(stem) )
    if type(branches) != type([]):
        raise Exception, "Branches is not a list as required, but a %s." % ( type(branch) )

    # Cut off must be set lower than standard 0.6, since stars will be counted as non-hits
    # Must allow for more variables in subline, 0.2 allows for [ * -> a -> b ] and [ * -> c -> b ] to be equal
    cutoff = 0.2
    num_results = 2

    candidates = get_close_matches(subline, branches, num_results, cutoff)

    # It no candidates came out of matching the subline, match entire lines
    # then remove the stems from resulting candidates
    if len(candidates) == 0:

        complete_branches = [ stem + b for b in branches ]

        complete_branch_candidates = get_close_matches( line, complete_branches, num_results, cutoff )

        candidates = [ candidate[stl:] for candidate in complete_branch_candidates ]
        
    return candidates


def add_to_tree( node, line, depth=None, branch=None, layout=None, debug=False ) :
    
    if branch and ( ( layout==None ) or ( depth==None ) ) :
        raise Exception, "When inserting a line into an existing branch, we need both depth and layout."

    parent = node
    n = len( line )
    
    if depth :
        tail_length = depth
    else : 
        tail_length = ( n - 1)

    if tail_length < 0 :
        raise Exception, "Tail length can never be less than zero(%i). Can not proceed inserting under node %s." % ( tail_length, parent )

    
    if branch and debug: 
        print "----------"
        print "Inserting line: ", line
        b = []
        for n in branch:
            b.append(n.label)
        print "Into branch: ", b
        print "Using layout: ",layout
    
    # Branch is not set, meaning that we can insert every item in the
    # line directly, after this, we are finished and can return
    if not branch :
        if debug:
            print "\tInserting line in its entirety. Starting with %s with tail_length %i" % ( line[0], tail_length )
            print "\tParent node is %s. " % parent
        for item in line :
            child = TreeNode( item, tail_length )
            parent.set_child( child ) 
            parent = child
            tail_length -= 1
        parent.increase_pc()
        
        if debug: print "\tNewly set parent node (%s)'s pattern count set to: %i. Has node count: %i." % ( parent, parent.pattern_count, parent.count )
        
        return

    
    if len(branch) != len(line) :
        raise Exception, "Branch's lenght(%i) it not equal to line length(%i). This is very wrong. " % ( len(branch), len(line) )
   
    for index, child in enumerate( branch ) :
        
        label = node.label

        if not child :
            if debug: print "Parent %s has children: " % parent, parent.get_children_details()
            raise Exception, "Child with label %s and tail length %s does not exist under parent %s. Impossible." % ( label, tail_length, parent )

        # Simplest case: 
        # Layout is 0 and child stays unchanged. Simply add to its node
        # count.
        if layout[index] == 0:
            if child.label != line[index] :
                raise Exception, "Layout says 0, but child label(%s) and line item(%s) differs. Not good. " % ( child.label, line[index])
            child.increase_nc()
            parent = child
            tail_length -= 1
            continue

        # Existing star case :
        # If layout shows a 2, then a star exists in the branch, and we
        # want to add hidden labels to the existing starnode. We must
        # look forwards into the layout to anticipate the value of the
        # following nodes, to make sure that the hidden label's
        # successors are correct. 
        if layout[index] == 2 :

            if child.label != '*' :
                raise Exception, "Layout claims node %s's label should be a star. It is not: " % child
            
            # If the remaining layout contains one or more 1's, it means that the
            # branch will change. To be able to update the existing
            # hidden labels, we must genereat both the previous and the
            # the new list of successors.
            
            prev_successors = child.calculate_successor( branch )
            if debug: 
                print "Layout = 2: Child %s get new hidden label: %s ." % ( child, line[index] )
                print "\tSuccessor generated: %s " % prev_successors

            remaining_layout = layout[index+1:]

            if 1 in remaining_layout : # successors will change
                new_successors = child.calculate_successor( branch, layout )
                if debug: 
                    print "\tRemaining layout is: ", remaining_layout
                    print "\tNew successor generated: %s " % new_successors
                child.set_hidden_label( line[index], prev_successors, new_successor=new_successors, debug=debug )
            else :
                child.set_hidden_label( line[index], prev_successors, debug=debug )

            if debug: 
                if child.calculate_no_labels() < 10 :
                    print "\tAfter adding hl, child %s has hidden label successors: " % child, child.hidden_labels.keys()
                else:
                    print "\tAfter adding hl, child %s has %s hidden label successors: " % ( child, child.calculate_no_labels() )
                print "Finished adding hidden label."
            parent = child
            tail_length -= 1

        # Replace node label
        # If layout says 1, it means that the node label should be
        # replaced by a star, and hidden labels added.

        if layout[index] == 1 :

            if child.label == line[index] :
                raise Exception, "Layout says 1, and claims child label(%s) and line item(%s) should be different..." % ( child.label, line[index] )

            if debug: print "Layout = 1: Child %s will be renamed ( new label: %s ) ." % ( child, line[index] )

            child.rename_node( line[index], branch, layout, debug=debug) # rename_node will update successors if needed
            
            if debug: 
                if child.calculate_no_labels() < 10 :
                    print "Renamed child %s has hidden labels: " % child, child.hidden_labels
                else :
                    print "Renamed child %s has hidden label keys: " % child, child.hidden_labels.keys()
                    
            parent = child
            tail_length -= 1
    
    # We have now gone trough all the items to be inserted into the
    # tree. We must now make sure to add a final pattern_count to the
    # last node inserted.
    parent.increase_pc()
    if debug:
        print "Finished inserting line. Tree under node %s so far: " % node
        branches = node.dump_tree( branch_count=False )
        bl = node.get_branch_labels( branches, details=True, debug=True )
        


if __name__ == "__main__":

    setup = True
    config_file = 'TreeStructure.config'
    pattern_file_name ='patterns'

    debug = False
    stages = False
    logfile = None
    patternfile = None
    progress = False
   
    argv = sys.argv[1:]

    try:                                
        opts, args = getopt.getopt(argv, "hdsrc:l:p:", ["help", "debug", "stages", "progress", "config="])
    except getopt.GetoptError:    
        usage()                     
        sys.exit(2)
    if ('-h' in argv ) or ( '--help' in argv ) :
        usage()
        sys.exit()
    if ( '-l' not in argv ) or ( '-p' not in argv ) :
        usage()
        sys.exit(2)

    for opt, arg in opts :
        if opt in ('-c','--config') :
            if not os.path.exists( arg ) :
                print "Config file '%s' does not exist." % arg
                sys.exit()
            print "Using config file: %s" % arg
            config_file = arg
        elif opt in ('-d', '--debug' ) :
            debug = True
        elif opt in ('-s', '--stages' ) :
            stages = True
        elif opt in ('-r', '--progress' ) :
            progress = True
        elif opt == '-l' :
            if not os.path.isfile( arg ) :
                print "Logfile '%s' does not exist." % arg
                sys.exit()
            logfile = arg
        elif opt == '-p' :
            if not os.path.isdir( arg ) :
                print "'%s' is not a valid directory." % arg
                sys.exit()
            path = arg
            if arg[-1] != '/':
                path = arg+'/'
            patternfile = path+pattern_file_name


    # Read config file
        
    config = ConfigParser.RawConfigParser()
    config.read( config_file )

    # Exceptions (or spam...)
    exceptions_file = config.get('Exceptions', 'exceptions_file')

    graph_node_threshold        = config.getint('Graph', 'graph_node_threshold')
    # Variables to control how lines are merged
    short_line          = config.getint('Variables', 'short_line')
    very_short_line     = config.getint('Variables', 'very_short_line')
    short_subline       = config.getint('Variables', 'short_subline')
    very_short_subline  = config.getint('Variables', 'very_short_subline')

    min_common_fraction         = config.getfloat('Variables', 'min_common_fraction')
    long_stem_fraction          = config.getfloat('Variables', 'long_stem_fraction')
    short_stem_common_fraction  = config.getfloat('Variables', 'short_stem_common_fraction')
    long_stem_common_fraction   = config.getfloat('Variables', 'long_stem_common_fraction')
    min_matches_required        = config.getfloat('Variables', 'min_matches_required')
    max_changes_allowed         = config.getfloat('Variables', 'max_changes_allowed')
    short_min_matches_required  = config.getfloat('Variables', 'short_min_matches_required') 
    short_max_changes_allowed   = config.getfloat('Variables', 'short_max_changes_allowed')

    hl_release_limit = config.getfloat('HiddenLabels', 'hl_release_limit')    

    # Start!

    main( )
