# -*- coding: utf-8 -*-

from pygraphviz import *
import libxml2, libxslt


# draw_tree2 outputs a png file instead of an svg. 
# Faster to draw, but not as pretty
def draw_tree2(root, file, threshold):
    file = file.replace(".svg", ".png")
    graph = AGraph(label="Testgraph",ranksep='0.2 equally')
    graph.node_attr['shape'] = 'box'
    graph.node_attr['style'] = 'filled'
    graph.node_attr['fillcolor'] = 'orange'
    
    id = root.nounce
    graph.add_node(id, label=root.label, fillcolor='red')
    draw_node(graph, root, id)
    save_graph(graph, file, threshold)


def draw_tree(root, file, threshold):
    graph = AGraph(label="Testgraph",ranksep='0.2 equally')
    graph.node_attr['shape'] = 'box'
    graph.node_attr['style'] = 'filled'
    graph.node_attr['fillcolor'] = 'orange'
    
    id = root.nounce
    graph.add_node(id, label=root.label, fillcolor='red')
    draw_node(graph, root, id)
    save_graph(graph, file, threshold)

def draw_node(graph, node, parent_id):
    
    children = node.children
    
    # If node is the last node, set color back to orange and return
    if len(children) == 0:
        p = graph.get_node(parent_id)
        #p.attr['fillcolor'] = 'orange'
        return
    
    for child in children:
        #id = (child.__repr__()).replace(' ','')
        id = child.nounce
        count = child.count
        #lab = "%s %s" % (child.label, child.nounce)
        lab = "%s" % child.label
        e_lab = str(count)
        
        #if child.hidden_labels:
        #   l = str( len(child.hidden_labels.keys()) )
        #   l = child.sum_hidden_labels()
        #   lab = "%s hl:%i" % ( str(child), l )
        #   lab = "%s" % child.label
        #if child.pattern_count:
        #    lab = lab + " PC: " + str(child.pattern_count)
      
        if child.pattern_count:
            graph.add_node(id, label=lab, shape='note', fillcolor='red')
        else:
            graph.add_node(id, label=lab, fillcolor='orange')
        
        graph.add_edge(parent_id, id, label=e_lab)
        draw_node(graph, child ,id)

    

def save_graph(graph, file, threshold):
    
    #print graph.string() # print to screen

    size = len( graph.nodes() )

    if size > threshold:
        print "\tGraph has %i nodes. Will not draw above %i nodes." % (size, threshold)
        return

    graph.layout('dot')
    graph.draw(file)
    #graph.draw(fname, format='ps', prog='dot') # neato

    if ".svg" in file:
        styledoc = libxml2.parseFile("notugly.xsl")
        style = libxslt.parseStylesheetDoc(styledoc)
        doc = libxml2.parseFile(file)
        result = style.applyStylesheet(doc, None)
        style.saveResultToFilename(file, result, 0)

    print "Graph saved to: ", file
    print "Number of nodes in graph: ",len(graph.nodes())
