# -*- coding: utf-8 -*-
import unittest
from TreeNode2 import TreeNode
from msg_extraction import extract_message
from tree_graph import draw_tree, draw_tree2
from TreeStructure import find_matching_branch, add_to_tree

#
# Unittests 
#

class Tests(unittest.TestCase):

    def print_tree( self, parent, title ) :

        print "\t=== Test: %s: " % title
        tree = parent.dump_tree()
        return_tree = []
        for branch in tree :
            c = branch.pop() # popping off the branch count
            tmp = []
            tmp2 = []
            for node in branch :
                tmp.append("%s(%i)(%i)" % ( node.label, node.count, node.pattern_count ) )
                tmp2.append("%s" % node)
            return_tree.append(tmp)
            print "\t\t",tmp2, c
        return return_tree


    def split_example( self ) :
        # parent -> A -> B -> E -> F
        # parent -> A -> B -> R -> S
        # parent -> A -> B -> R -> S
    
        parent = TreeNode('parent',4)
        A = TreeNode('A',3)
        B = TreeNode('B',2)
        E = TreeNode('E',1)
        F = TreeNode('F',0)
        R = TreeNode('R',1)
        S = TreeNode('S',0)

        # First line
        parent.set_child(A)
        A.set_child(B)
        B.set_child(E)
        E.set_child(F)
        F.increase_pc()

        # Second line
        parent.increase_nc()
        A.increase_nc()
        B.increase_nc()
        B.set_child(R)
        R.set_child(S)
        S.increase_pc()

        # Third line
        parent.increase_nc()
        A.increase_nc()
        B.increase_nc()
        R.increase_nc()
        S.increase_nc()
        S.increase_pc()

        initial = self.print_tree( parent, "Split example" )

        initial_corr =  [   [ 'A(3)(0)','B(3)(0)','E(1)(0)','F(1)(1)' ],\
                            [ 'A(3)(0)','B(3)(0)','R(2)(0)','S(2)(2)' ] ]

        self.assertEqual( initial, initial_corr )

        return parent, A, B, E, F, R, S


    def split_rename_example( self ) :

        # parent -> A -> B
        # parent -> A -> B -> E -> F
        # parent -> A -> B -> R -> S
        # parent -> A -> B -> R -> S

        # Create all the nodes we need
        parent = TreeNode( 'parent',4 )
        A = TreeNode('A',3)
        B = TreeNode('B',2)
        E = TreeNode('E',1)
        F = TreeNode('F',0)
        R = TreeNode('R',1)
        S = TreeNode('S',0)

        # First line
        parent.set_child(A)
        A.set_child(B)
        B.increase_pc()
        # Second line
        parent.increase_nc()
        A.increase_nc()
        B.increase_nc()
        B.set_child(E)
        E.set_child(F)
        F.increase_pc()
        # Third line
        parent.increase_nc()
        A.increase_nc()
        B.increase_nc()
        B.set_child(R)
        R.set_child(S)
        S.increase_pc()
        # Fourth line
        parent.increase_nc()
        A.increase_nc()
        B.increase_nc()
        R.increase_nc()
        S.increase_nc()
        S.increase_pc()

        res = self.print_tree( parent, "Split and rename example" )

        # parent -> A -> B
        # parent -> A -> B -> E -> F
        # parent -> A -> B -> R -> S
        # parent -> A -> B -> R -> S
        

        corr =  [   [ 'A(4)(0)','B(4)(1)' ],\
                    [ 'A(4)(0)','B(4)(1)','E(1)(0)','F(1)(1)' ],\
                    [ 'A(4)(0)','B(4)(1)','R(2)(0)','S(2)(2)' ]
                ]
        self.assertEqual( res,corr )
                
        return parent, A, B, E, F, R, S

    def test_create_node( self ) :
        node = TreeNode('label',5)
        self.assertEqual('label',node.label)

    def test_create_child( self ) :
        node = TreeNode('with_children',5)
        child1 = TreeNode('child1',4)
        child2 = TreeNode('child2',4)
        node.set_child(child1)
        node.increase_nc() # correct node count
        node.set_child(child2)
        l = [child1,child2]
        self.assertEqual(node.children, l)

    def test_remove_child( self ) :
        node = TreeNode('parent', 5)
        child1 = TreeNode('child1',4)
        child2 = TreeNode('child2',4)
        node.set_child(child1)
        node.increase_nc() # correct node count
        node.set_child(child2)
        node.remove_child(child1)
        self.assertEqual(node.children, [child2])

    def test_del_child( self ):
        node = TreeNode('for_deletion',1)
        child = TreeNode('child',0)
        node.set_child(child)
        node.remove_child(child)
        child.destruct()

    def test_simple_rename_node( self ) :
        
        print "\t**** Simple rename  ****"
       
        # parent -> to_be_renamed (x3)
        # parent -> *

        parent = TreeNode('parent',1, node_count=3 )
        node = TreeNode('to_be_renamed', 0, node_count=3, pattern_count=3 )
        
        parent.set_child(node)
        
        nc_before = node.count

        parent.increase_nc()
        node.rename_node( 'new_label', [ parent, node ], [ 0, 0 ] ) # branch, layout
        node.increase_pc()
        
        self.assertEqual(parent.count, 4 )
        self.assertEqual(node.count, 4 )
        self.assertEqual(node.label, '*')
        self.assertEqual(node.count, nc_before+1)
        self.assertEqual(node.hidden_labels, { '' : {'to_be_renamed':3, 'new_label':1} } )

        res_tree = self.print_tree( parent, "Simple rename node, result" )
        
        corr_tree =  [  [ '*(4)(4)' ] ]
        self.assertEqual( res_tree, corr_tree )
        
    
    def test_simple_split( self ) :

        print "\t**** Simple split  ****"
        
        # parent -> A -> B -> E -> F
        # parent -> A -> B -> R -> S
        # parent -> A -> B -> R -> S

        parent, A, B, E, F, R, S = self.split_example()

        # Appending:
        # parent -> A -> B -> R -> S -> T
        T = TreeNode('T', 0)
        parent.increase_nc()
        A.increase_nc()
        B.increase_nc()
        R.increase_nc()
        S.increase_nc()
        S.set_child(T)
        T.increase_pc()

        A.split_branch( [ B, R, S ], 2 ) # Note: branch count = 2

        res_tree = self.print_tree( parent, "Simple split, result" )
        
        corr_tree =  [  [ 'A(2)(0)','B(2)(0)','R(2)(0)','S(2)(2)' ],\
                        [ 'A(2)(0)','B(2)(0)','E(1)(0)','F(1)(1)' ],\
                        [ 'A(2)(0)','B(2)(0)','R(1)(0)','S(1)(0)','T(1)(1)' ] ]
        self.assertEqual( res_tree, corr_tree )
        
        self.assertEqual( parent.count, 4 )


    def test_simple_split2( self ) :

        print "\t**** Simple slit 2  ****"

        # parent -> A -> B -> E -> F
        # parent -> A -> B -> R -> S
        # parent -> A -> B -> R -> S
        # parent -> > -> B -> R -> S -> T

        parent, A, B, E, F, R, S = self.split_example()
        
        A.split_branch( [ B, R, S], 2 ) # Note: branch count = 2

        res_tree = self.print_tree( parent, "Simple split 2, result" )
        corr_tree =  [  [ 'A(2)(0)','B(2)(0)','R(2)(0)','S(2)(2)' ],\
                        [ 'A(1)(0)','B(1)(0)','E(1)(0)','F(1)(1)' ] ]
        self.assertEqual( res_tree, corr_tree )

        self.assertEqual( parent.count, 3 ) 


    def test_complex_rename_node( self ) :
        
        print "\n**** Complex rename  ****"
       
        # parent -> A -> B
        # parent -> A -> B -> E -> F
        # parent -> A -> B -> R -> S
        # parent -> A -> B -> R -> S

        parent, A, B, E, F, R, S = self.split_rename_example()
        
        # insert : parent -> * -> B -> R -> S

        # Fifth line
        parent.increase_nc()
        A.rename_node( 'kake', [ parent, A, B, R, S ], [ 0, 1, 0, 0, 0 ] )
        B.increase_nc()
        R.increase_nc()
        S.increase_nc()
        S.increase_pc()
        
        self.assertEqual( parent.count, 5 )

        res_tree = self.print_tree( parent, "Complex rename node, result" )
        corr_tree =  [  [ '*(3)(0)','B(3)(0)','R(3)(0)','S(3)(3)' ],\
                        [ 'A(2)(0)','B(2)(1)'],\
                        [ 'A(2)(0)','B(2)(1)','E(1)(0)','F(1)(1)' ] ]
        self.assertEqual( res_tree, corr_tree )
        
        
    def test_remove_stars( self ) :
        
        print "\**** Remove stars  ****"
        
        # Create all the nodes we need
        # parent -> *(pepper) -> A -> *(kake) -> B -> *(kake)
        # parent -> *(+mann) -> A -> *(+blapp) -> C -> *()
        # parent -> * -> A -> B -> *(C) -> *
        
        parent = TreeNode('parent',5)
        c1 = TreeNode('*',4)
        c1.set_hidden_label('pepper','A,B', increase_count=False)
        g1 = TreeNode('A',3)
        gg1 = TreeNode('*',2)
        gg1.set_hidden_label('kake','B', increase_count=False)
        ggg1 = TreeNode('B',1)
        gggg1 = TreeNode('*',0)
        gggg1.set_hidden_label('kake','####', increase_count=False)
        
        # Inserting first line:
        parent.set_child(c1)
        c1.set_child(g1)
        g1.set_child(gg1)
        gg1.set_child(ggg1)
        ggg1.set_child(gggg1)
        gggg1.increase_pc()

        ggg12 = TreeNode('C',1)
        gggg12 = TreeNode('*',0)

        # Inserting second line:
        parent.increase_nc()
        c1.set_hidden_label('mann','A,B')
        g1.increase_nc()
        gg1.set_hidden_label('blapp','C')
        gg1.set_child(ggg12)
        ggg12.set_child(gggg12)
        gggg12.increase_pc()

        c2 = TreeNode('*',4)
        g2 = TreeNode('A',7)
        gg2 = TreeNode('B',2)
        ggg2 = TreeNode('*',1)
        ggg2.set_hidden_label('C','', increase_count=False)
        gggg2 = TreeNode('*',0)
      
        # Inserting third line:
        parent.increase_nc() 
        parent.set_child(c2, ignore_duplicates=True)
        c2.set_child(g2)
        g2.set_child(gg2)
        gg2.set_child(ggg2)
        ggg2.set_child(gggg2)
        gggg2.increase_pc()


        initial = self.print_tree( parent, "remove stars, initial tree" )
    
        print "Removing stars:"
        parent.remove_stars()

        self.assertEqual(parent.children, [g1,g2])
        self.assertEqual(g1.children, [ggg1,ggg12])
        self.assertEqual(ggg1.children, [])
        self.assertEqual(gg2.children, [])
        # Node A should inherit labels from its starchild
        self.assertEqual(g1.hidden_labels, { 'B':{'kake':1}, 'C':{'blapp':1} } )
        self.assertEqual(ggg2.hidden_labels, { '':{'C':1} } )
        
        # Correct result should be:
        # ['A(2)(0)', 'B(1)(0)'] 1
        # ['A(2)(0)', 'C(1)(0)'] 1
        # ['A(1)(0)', 'B(1)(0)'] 1
       
        result = self.print_tree( parent, "remove stars, result" )
        # label, count, pc
        self.assertEqual( result, [ ['A(2)(0)','B(1)(1)'], ['A(2)(0)','C(1)(1)'], ['A(1)(0)','B(1)(1)'] ] )

    def test_merge( self ) :
        
        print "\n**** Merging  ****"

        # Initial tree
        # parent -> A -> B -> G
        # parent -> A -> B
        # parent -> A -> B -> D -> E

        parent = TreeNode('parent',5)
        # First line
        A1 = TreeNode('A',2)
        B1 = TreeNode('B',1)
        G = TreeNode('G',0)
        # Second line
        A2 = TreeNode('A',1)
        B2 = TreeNode('B',0)
        # Third line
        A3 = TreeNode('A',3)
        B3 = TreeNode('B',2)
        D = TreeNode('D',1)
        E = TreeNode('E',0)
        
        # Insert first line
        parent.set_child(A1, ignore_duplicates=True)
        A1.set_child(B1, ignore_duplicates=True)
        B1.set_child(G, ignore_duplicates=True)
        G.increase_pc()
        # Insert second line
        parent.increase_nc()
        parent.set_child(A2, ignore_duplicates=True)
        A2.set_child(B2, ignore_duplicates=True)
        B2.increase_pc()
        # Inserting third line
        parent.increase_nc()
        parent.set_child(A3, ignore_duplicates=True)
        A3.set_child(B3, ignore_duplicates=True)
        B3.set_child(D, ignore_duplicates=True)
        D.set_child(E, ignore_duplicates=True)
        E.increase_pc()

        initial = self.print_tree( parent, "Merge, initial tree" )

        # And then: merge!
        parent.merge()

        result = self.print_tree( parent, "Merge, result" )
        corr = [    ['A(3)(0)','B(3)(1)'],\
                    ['A(3)(0)','B(3)(1)','G(1)(1)'],\
                    ['A(3)(0)','B(3)(1)','D(1)(0)','E(1)(1)'] ]
        self.assertEqual( result, corr )

        self.assertEqual( len(parent.children), 1 )
        a = parent.children[0]
        self.assertEqual( a.label, 'A')
        self.assertEqual( a.count, 3 )
        self.assertEqual( len(a.children), 1 )
        b = a.children[0]
        self.assertEqual( b.label, 'B')
        self.assertEqual( b.count, 3 )
        self.assertEqual( b.pattern_count, 1 )
        self.assertEqual( len(b.children), 2 )
        g = b.children[0]
        d = b.children[1]
        self.assertEqual( g.count, 1 )
        self.assertEqual( d.count, 1 )
    
    
    def test_simple_release_hls( self ) :
        print "**** Simple release of hls: ****"
        # Initial tree:
        # parent -> A
        # parent -> A (new a)
        # parent -> *(B) (on new a)

        parent = TreeNode('parent',2)
        A = TreeNode('A',0)
        star = TreeNode('A',1)
    
        # First line
        parent.set_child(A)
        A.increase_pc()

        # Second line
        parent.increase_nc()
        parent.set_child(star)
        star.increase_pc()

        # Third line
        parent.increase_nc()
        star.rename_node('B',[ parent, star ], [ 0, 1 ])
        star.increase_pc()
       
        parent.remove_stars()

        first_res = self.print_tree( parent, "simple_release, first results" )
        corr_first = [ ['A(1)(1)'] ]

        self.assertEqual(parent.hidden_labels, {'' : {'A':1, 'B':1} } )
        
        parent.merge()
        
        parent.release_hls( 0.2 ) # hl_release_limit

        result = self.print_tree( parent, "simple_release, results" ) 
        corr = [ ['A(2)(2)'],['B(1)(1)'] ]

        self.assertEqual(len(parent.hidden_labels),0)

    def ztest_dump_tree( self ) :
        print "\n**** Tree dump test ****"
        
        parent = TreeNode('parent',1)
        child = TreeNode('child',0)
        parent.set_child(child)
        child.increase_pc()

        parent.increase_nc()
        parent.increase_pc(1)

        print "Parent has children: ", parent.children

        result = self.print_tree( parent, "dump tree results") 
        self.assertEqual(result, [['child(1)(1)']])
        

    def test_release_hls( self ) :
        print "\n**** Release hls ****"
        
        # Initial tree:
        # parent -> B -> R
        # parent -> B -> R -> q
        # parent -> B -> R -> q
        # parent -> C -> R -> q 

        parent = TreeNode('parent',3)
        B = TreeNode('B',1)
        R2 = TreeNode('R',0)

        star = TreeNode('B',2)
        R = TreeNode('R',1)
        q = TreeNode('q',0)
        
        # First line
        parent.set_child(B)
        B.set_child(R2)
        R2.increase_pc()

        # Second line
        parent.increase_nc()
        parent.set_child(star)        
        star.set_child(R)
        R.set_child(q)
        q.increase_pc()
        
        # Third line
        parent.increase_nc()
        star.increase_nc()
        R.increase_nc()
        q.increase_nc()
        q.increase_pc()

        # Fourth line
        parent.increase_nc()
        star.rename_node('C', [ parent, star, R, q ], [ 0, 1, 0, 0 ] )
        R.increase_nc()
        q.increase_nc()
        q.increase_pc()

        self.assertEqual(star.hidden_labels, { 'R,q': {'B':2, 'C':1} } )

        init = self.print_tree( parent, "release hls, initial" )
        corr_init = [   [ 'B(1)(0)', 'R(1)(1)' ],\
                        [ '*(3)(0)', 'R(3)(0)', 'q(3)(3)' ] ]
    
        self.assertEqual( init, corr_init )

        parent.remove_stars()

        self.assertEqual( parent.hidden_labels, { 'R,q': {'B':2, 'C':1} } )

        res_stars = self.print_tree( parent, "release hls, removed stars" )
        
        parent.merge()

        res_merge = self.print_tree( parent, "release hls, merged" )
        corr_merge = [   [ 'B(1)(0)', 'R(1)(1)' ],\
                         [ 'R(3)(0)', 'q(3)(3)' ] ]
        self.assertEqual( res_merge, corr_merge )

        parent.release_hls( 0.2 ) # hl_release_limit

        result = self.print_tree( parent, "release hls, results" )
        corr = [    [ 'B(3)(0)','R(3)(1)'],\
                    [ 'B(3)(0)','R(3)(1)','q(2)(2)' ],\
                    [ 'C(1)(0)','R(1)(0)','q(1)(1)' ] ]
        self.assertEqual( result, corr )
        self.assertEqual( parent.hidden_labels, {} )
    
    def test_release_hls2( self ) :
        
        print "**** Release hls 2****"

        # Initial tree:
        # parent -> A
        # parent -> C
        # parent -> A -> R 

        parent = TreeNode('parent',1)
        A = TreeNode('A',0)
        G = TreeNode('C',0)
        
        star = TreeNode('A',1)
        R = TreeNode('R',0)

        star2 = TreeNode('*', 3, node_count=2)
        star2.hidden_labels['K,L,M'] = {'A':2}
        K = TreeNode('K', 2, node_count=2)
        L = TreeNode('L', 1, node_count=2)
        M = TreeNode('M', 0, node_count=2)
        

        # First and second line
        parent.set_child(A)
        A.increase_pc()
        parent.increase_nc()
        parent.set_child(G)
        G.increase_pc()
        # Third line
        parent.increase_nc()
        parent.set_child(star)
        star.set_child(R)
        R.increase_pc()
        # Fourth line
        parent.increase_nc()
        star.rename_node('C', [ parent, star, R ], [ 0, 1, 0 ] )
        R.increase_nc()
        R.increase_pc()
        # Fifth line
        parent.increase_nc(2)
        parent.set_child(star2)
        star2.set_child(K)
        K.set_child(L)
        L.set_child(M)
        M.increase_pc(2)
        
        initial = self.print_tree( parent, "release hls 2, initial" )

       # draw_tree2( parent, 'filename' )

        parent.remove_stars()
        print "Merge:"
        parent.merge()
        print "Release hls:"
        parent.release_hls( 0.2 ) # hl_release_limit

        result = self.print_tree( parent, "release hls 2, results" )

        self.assertEqual(parent.children, [A,G])
        self.assertEqual(A.count, 4)
        self.assertEqual(A.pattern_count, 1)
        self.assertEqual(len(A.children), 2)
        R1 = A.children[0]
        self.assertEqual(R1.label, 'R')
        self.assertEqual(R1.count, 1)
        self.assertEqual(R1.pattern_count, 1)
        K1 = A.children[1]
        self.assertEqual(K1.label, 'K')
        self.assertEqual(K1.count, 2)

        self.assertEqual(G.count,2)
        self.assertEqual(G.pattern_count, 1)
        self.assertEqual(len(G.children), 1)
        R2 = G.children[0]
        self.assertEqual(R2.label, 'R')
        self.assertEqual(R2.count, 1)
        self.assertEqual(R2.pattern_count, 1)


    def test_successors_renaming( self ) :
        
        print "\t**** Successors renaming ****"

        # Initial tree:
        # parent -> E -> B -> C
        # parent -> F -> B -> C
        # parent -> F -> B -> D

        parent = TreeNode('parent',3)
        star = TreeNode('E',2)
        B = TreeNode('B',1)
        C = TreeNode('C',0)

        # First line
        parent.set_child(star)
        star.set_child(B)
        B.set_child(C)
        C.increase_pc()
        
        # Second line
        parent.increase_nc()
        star.rename_node( 'F', [ parent, star, B, C ], [ 0, 1, 0, 0 ] )
        B.increase_nc()
        C.increase_nc()
        C.increase_pc()
    
        inital = self.print_tree( parent, "Successors renaming, before renaming" )

        # Third line
        parent.increase_nc()
        star.set_hidden_label('F','B,C','B')
        B.increase_nc()
        C.rename_node('D', [ parent, star, B, C ], [ 0, 0, 0, 1 ] )
        C.increase_pc()

        result = self.print_tree( parent, "Successors renaming, results" )

        self.assertEqual(star.hidden_labels, {'B' : {'E':1, 'F':2} })
        self.assertEqual(C.hidden_labels, {'': {'C':2, 'D':1} })

    def ztest_merging_hls( self ) :
        
        print "**** Merging of hls ****"

        # Initial tree:
        # parent -> A -> S 
        # parent -> A -> R -> D
        # parent -> A -> *(R,S) -> D
        # parent -> A -> *(R,S,T) -> D

        parent = TreeNode('parent',3)
        A = TreeNode('A',1)
        S = TreeNode('S',0)
        A2 = TreeNode('A',2)
        star = TreeNode('R',1)
        D = TreeNode('D',0)

        # First line
        parent.set_child(A)
        A.set_child(S)
        S.increase_pc()
        # Second line
        parent.increase_nc()
        parent.set_child(A2)
        A2.set_child(star)
        star.set_child(D)
        D.increase_pc()
        # Third line
        parent.increase_nc()
        A2.increase_nc()
        star.rename_node('S', [parent, A2, star, D], [0, 0, 1, 0] )
        D.increase_nc()
        D.increase_pc()
        # Fourth line
        parent.increase_nc()
        A2.increase_nc()
        star.set_hidden_label('T','D')
        D.increase_nc()
        D.increase_pc()


        parent.remove_stars()
        self.assertEqual(A2.hidden_labels, {'D' : {'R':1 ,'S':1, 'T':1} })
        
        parent.merge()
        self.assertEqual(len(parent.children), 1)

        #draw_tree2( parent, 'filename' )
        
        parent.release_hls( 0.2 ) # hl_release_limit

        result = self.print_tree( parent, "Merging of hls, results" )
        corr = [    [ 'A(4)(0)','S(2)(1)'],\
                    [ 'A(4)(0)','S(2)(1)','D(1)(1)'],\
                    [ 'A(4)(0)','D(2)(2)'] ]
        self.assertEqual( result, corr )

        self.assertEqual( len( A2.children ), 2 )
        self.assertEqual( A2.hidden_labels, {'D' : {'R':1 ,'T':1} })

    def test_finding_brances( self ) :
        
        print "\n**** Finding branches ****"

        # Initial tree:
        # parent -> B -> E -> F
        # parent -> C -> T
        # parent -> D -> R -> S
        # parent -> G -> H -> I -> J

        parent = TreeNode('parent', 4)
        B = TreeNode('B',2)
        C = TreeNode('C',1)
        D = TreeNode('D',2)
        G = TreeNode('G',3)
        
        E = TreeNode('E',1)
        F = TreeNode('F',0)
        T = TreeNode('T',0)
        R = TreeNode('R',1)
        S = TreeNode('S',0)
        H = TreeNode('H',2)
        I = TreeNode('I',1)
        J = TreeNode('J',0)
        

        # First line
        parent.set_child(B)
        B.set_child(E)
        E.set_child(F)
        F.increase_pc()
        # Second line
        parent.increase_nc()
        parent.set_child(C)
        C.set_child(T)
        T.increase_pc()
        # Third line
        parent.increase_nc()
        parent.set_child(D)
        D.set_child(R)
        R.set_child(S)
        S.increase_pc()
        # Fourth line
        parent.increase_nc()
        parent.set_child(G)
        G.set_child(H)
        H.set_child(I)
        I.set_child(J)
        J.increase_pc()

        initial = self.print_tree( parent, "Find branches, initial tree" )

        print "Now, look for any branch with length 3"
        b1 = parent.dump_tree( find_matching=True, depth=3 ) 

        print "Now, look for branches with length 5"
        b2 = parent.dump_tree( find_matching=True, depth=5)
        
        branches = []
        for similar in b1 :
            bl1 = []
            for node in similar :
                bl1.append(node.label)
            branches.append(bl1)
            print "b1 branch found: ", bl1 

        branches2 = []
        for similar in b2 :
            bl2 = []
            for node in similar :
                bl2.append(node.label)
            branches.append(bl2)
            print "b2 branch found: ", bl2 


        self.assertEqual(branches, [ ['B','E','F'], ['D','R','S'] ] )
        self.assertEqual(b2, [] )

    def test_merging_with_children( self ) :
        
        print "\n**** Merging with children ****"

        # Initial tree:
        # parent -> B -> H -> S
        # parent -> B -> G -> R
        # parent -> B -> F
        # parent -> B -> E

        parent = TreeNode('parent', 3)
        B = TreeNode('B',2)
        H = TreeNode('H',1)
        S = TreeNode('S',0)
        G = TreeNode('G',1)
        R = TreeNode('R',0)
        
        B2 = TreeNode('B',1)
        F = TreeNode('F',0)
        E = TreeNode('E',0)

        # First line
        parent.set_child(B)
        B.set_child(H)
        H.set_child(S)
        S.increase_pc()
        # Second line
        parent.increase_nc()
        B.increase_nc()
        B.set_child(G)
        G.set_child(R)
        R.increase_pc()
        # Third line
        parent.increase_nc()
        parent.set_child(B2)
        B2.set_child(F)
        F.increase_pc()
        # Fourth line
        parent.increase_nc()
        B2.increase_nc()
        B2.set_child(E)
        E.increase_pc()

        initial = self.print_tree( parent, "Merging with children, initial" )

        parent.merge()

        result = self.print_tree( parent, "Merging with children, result" )

        self.assertEqual( parent.children, [B] )
        self.assertEqual( len( B.children ), 4 )
        self.assertEqual( B.children, [ H,G,F,E ] )


    def test_merging_with_children2( self ) :
        
        print "\n**** Merging with children2 ****"

        # Initial tree:
        # parent -> A -> B -> C
        # parent -> A -> B -> C -> D

        parent = TreeNode('parent', 3, node_count=3)
        A = TreeNode('A',2, node_count=3 )
        B = TreeNode('B',1, node_count=3 )
        C = TreeNode('C',0, node_count=3 )
        A2 = TreeNode('A',3)
        B2 = TreeNode('B',2)
        C2 = TreeNode('C',1)
        D2 = TreeNode('D',0)
       
        # First line
        parent.set_child(A)
        A.set_child(B)
        B.set_child(C)
        C.increase_pc(3)

        # Second line
        parent.increase_nc()
        parent.set_child(A2)
        A2.set_child(B2)
        B2.set_child(C2)
        C2.set_child(D2)
        D2.increase_pc()

        initial = self.print_tree( parent, "Merging with children 2, initial" )

        self.assertEqual( initial, [ ['A(3)(0)','B(3)(0)','C(3)(3)'], ['A(1)(0)','B(1)(0)','C(1)(0)','D(1)(1)'] ] )
        
        parent.merge()
        
        result = self.print_tree( parent, "Merging with children 2, result" )
        self.assertEqual( result, [ ['A(4)(0)','B(4)(0)','C(4)(3)'], ['A(4)(0)','B(4)(0)','C(4)(3)','D(1)(1)'] ] )

    def test_merging_with_children3( self ) :
        
        print "\n**** Merging with children 3****"

        root = TreeNode('parent', 0, node_count=0)

        # Initial tree:
        lines = [   'A C D E G H',\
                    'A C D F R S',\
                    'A kake C D E G H',\
                    'A deig C D E G H',\
                    'A kake C D F R S',\
                    'A mann C D F R S' 
                ]
                    #'A sol C D arium E G H',\
                    #'A strand C D arium E G H' ]

        for line in lines :
            msg = line.split( )
        
            node, line, tail_length, branch, layout = matching_branch = find_matching_branch( root, msg )
            add_to_tree( node, line, depth=tail_length, branch=branch, layout=layout )
        
        print "Result after insert: "
        r0 = self.print_tree( root, "Merging with children 3, res: after insert")
        c0 = [  [ 'A(2)(0)', 'C(2)(0)', 'D(2)(0)', 'E(1)(0)', 'G(1)(0)', 'H(1)(1)' ],\
                [ 'A(2)(0)', 'C(2)(0)', 'D(2)(0)', 'F(1)(0)', 'R(1)(0)', 'S(1)(1)' ],\
                [ 'A(4)(0)', '*(4)(0)', 'C(4)(0)', 'D(4)(0)', 'F(2)(0)', 'R(2)(0)', 'S(2)(2)' ],\
                [ 'A(4)(0)', '*(4)(0)', 'C(4)(0)', 'D(4)(0)', 'E(2)(0)', 'G(2)(0)', 'H(2)(2)' ] ]
        self.assertEqual(r0, c0)
     
        root.remove_stars()
        
        r1 = self.print_tree( root, "Merging with children 3, res: without stars" )
        c1 = [  [ 'A(2)(0)', 'C(2)(0)', 'D(2)(0)', 'E(1)(0)', 'G(1)(0)', 'H(1)(1)'],\
                [ 'A(2)(0)', 'C(2)(0)', 'D(2)(0)', 'F(1)(0)', 'R(1)(0)', 'S(1)(1)'],\
                [ 'A(4)(0)', 'C(4)(0)', 'D(4)(0)', 'F(2)(0)', 'R(2)(0)', 'S(2)(2)'],\
                [ 'A(4)(0)', 'C(4)(0)', 'D(4)(0)', 'E(2)(0)', 'G(2)(0)', 'H(2)(2)'] ]
        self.assertEqual(r1, c1)

        root.merge()
    
        r2 = self.print_tree( root, "Merging with children 3, result" )
        c2 = [      [ 'A(6)(0)', 'C(6)(0)', 'D(6)(0)', 'E(3)(0)', 'G(3)(0)' 'H(3)(3)' ],\
                    [ 'A(6)(0)', 'C(6)(0)', 'D(6)(0)', 'F(3)(0)', 'R(3)(0)' 'S(3)(3)' ] ]

        self.assertEqual(r1, c1)
        A_now = root.children[0]
        self.assertEqual(A_now.hidden_labels, { 'C,D,F,R,S':{ 'kake':1, 'mann':1 }, 'C,D,E,G,H':{ 'kake': 1, 'deig':1 } } )

    def test_inserting_identical( self ) :
        
        print "\n**** Inserting identical lines ****"
        
        root = TreeNode( 'root', 0, node_count=0 )
        lines = [   ['mimedefang.pl', 'filter_sender', 'Internally', 'whitelisted'],\
                    ['mimedefang.pl', 'filter_sender', 'Internally', 'whitelisted']]

        for msg in lines :

            node, line, tail_length, branch, layout = matching_branch = find_matching_branch( root, msg )
            
            if node == None:  # This is added in the code
                continue
            
            add_to_tree( node, line, depth=tail_length, branch=branch, layout=layout )

        print "Result after insert: "
        r0 = self.print_tree( root, "Merging with children 3, res: after insert")
        c0 = [  [ 'mimedefang.pl(2)(0)', 'filter_sender(2)(0)', 'Internally(2)(0)', 'whitelisted(2)(2)' ] ]
        self.assertEqual(r0, c0)



    def test_split_hidden_labels( self ) :

        print "\n**** Splitting of hidden labels ****"
    
        root = TreeNode( 'root', 0, node_count=0 )

        # if first item to be inserted is different from the rest of the
        # tree, the line will be inserted in its entirety! ( because of process names.. )
        lines = [ "A B kake D E F G", "A B mann D E F G", "A B fjas F G H I", "A B tull F G H I", "A blah kake F G H I" ]
        
        for line in lines :
            msg = line.split( )
        
            node, line, tail_length, branch, layout = matching_branch = find_matching_branch( root, msg )
            add_to_tree( node, line, depth=tail_length, branch=branch, layout=layout )

        r0 = self.print_tree( root, "Splitting of hls, res: after insert")
        c0 = [  [ 'A(5)(0)', '*(3)(0)', '*(3)(0)', 'F(3)(0)', 'G(3)(0)', 'H(3)(0)', 'I(3)(3)' ],\
                [ 'A(5)(0)', 'B(2)(0)', '*(2)(0)', 'D(2)(0)', 'E(2)(0)', 'F(2)(0)', 'G(2)(2)' ] ]
        self.assertEqual( r0, c0 )
        
        A = root.get_child('A')
        B = A.get_child('B')
        star1 = A.get_child('*')
        star2 = star1.get_child('*')
        star3 = B.get_child('*')

        self.assertEqual( star1.hidden_labels, {'F,G,H,I' : { 'B':2, 'blah':1 } } )
        self.assertEqual( star2.hidden_labels, {'F,G,H,I' : { 'fjas':1 ,'tull':1, 'kake':1 } } )
        self.assertEqual( star3.hidden_labels, {'D,E,F,G' : { 'kake':1, 'mann':1 } } )


    def test_successor_calculation( self ) :

        A = TreeNode('A',0)
        B = TreeNode('*',1)
        C = TreeNode('C',2)
        D = TreeNode('*',3)
        E = TreeNode('E',4)
        F = TreeNode('F',5) # to be replaced

        branch = [ A, B, C, D, E, F ]

        old = A.calculate_successor( branch )
        new = A.calculate_successor( branch, layout=[0, 2, 0, 2, 0, 1]  )

        self.assertEqual( old, 'C,E,F')
        self.assertEqual( new, 'C,E' )
    
    def test_updating_labels( self ) :
        
        print "\n**** Updating hidden labels ****"

        # parent -> A -> B -> C -> D
        # parent -> *(R) -> B -> C -> D
        # parent -> *(S) -> B -> *(Q) -> D

        parent = TreeNode('parent',4)
        star1 = TreeNode('A',3)
        B = TreeNode('B',2)
        star2 = TreeNode('C',1) 
        D = TreeNode('D',0)

        # First line
        parent.set_child(star1)
        star1.set_child(B)
        B.set_child(star2)
        star2.set_child(D)
        D.increase_pc()

        # Second line
        parent.increase_nc()
        star1.rename_node('R', [ star1, B, star2, D ], [ 1, 0, 0, 0 ] ) 
        B.increase_nc()
        star2.increase_nc()
        D.increase_nc()
        D.increase_pc()

        # Third line
        parent.increase_nc()
        star1.set_hidden_label( 'S','B,C,D','B,D' )
        B.increase_nc()
        star2.rename_node('Q', [ star1, B, star2, D ], [ 2, 0, 1, 0 ] )
        D.increase_nc()
        D.increase_pc()

        r0 = self.print_tree( parent, "Updating hidden labels, initial" )

        self.assertEqual( star1.hidden_labels, { 'B,D' : { 'A':1, 'R':1, 'S':1 } } )
        self.assertEqual( star2.hidden_labels, { 'D': { 'C':2, 'Q':1 } } )


    # TODO: Make me work. Please!
    def test_complex_release( self ) :
        
        print "\n**** Complex release ****"

        # parent -> A -> *(S,T) -> D
        # parent -> A -> *(S,T) -> D -> K
        # parent -> A -> S -> D -> K
        # parent -> A -> S -> D -> R
        
        parent = TreeNode('parent', 4, node_count=2)
        A = TreeNode('A', 2, node_count=2)
        S = TreeNode('*', 1, node_count=2)
        S.hidden_labels = {'D' : { 'S':1, 'T':1 } }
        D = TreeNode('D', 0, node_count=2, pattern_count=2)

        S2 = TreeNode('*', 2, node_count=2)
        S2.hidden_labels = { 'D,K' : { 'S':1, 'T':1 } }
        D2 = TreeNode('D', 1, node_count=2)
        K2 = TreeNode('K', 0, node_count=2, pattern_count=2)

        S3 = TreeNode('S', 2 )
        D3 = TreeNode('D', 1 )
        K3 = TreeNode('K', 0, pattern_count=1 )
        R3 = TreeNode('R', 0, pattern_count=1 )

        # Juks og fanteri
        
        # First line
        parent.set_child(A)
        A.set_child(S)
        S.set_child(D)
        
        # Second line
        parent.increase_nc( 2 )
        A.increase_nc( 2 )
        A.set_child( S2 )
        S2.set_child( D2 )
        D2.set_child( K2 )
        
        init = self.print_tree( parent, "Complex release, after second line" )

        # Third line
        parent.increase_nc()
        A.increase_nc()
        A.set_child( S3 )
        S3.set_child( D3 )
        D3.set_child( K3 )

        # Fourth line
        parent.increase_nc()
        A.increase_nc()
        S3.increase_nc()
        D3.increase_nc()
        D3.set_child( R3 )
        

        init = self.print_tree( parent, "Complex release, initial" )
        c_init = [  ['A(6)(0)', '*(2)(0)', 'D(2)(2)'],\
                    ['A(6)(0)', '*(2)(0)', 'D(2)(0)', 'K(2)(2)'],\
                    ['A(6)(0)', 'S(2)(0)', 'D(2)(0)', 'K(1)(1)'],\
                    ['A(6)(0)', 'S(2)(0)', 'D(2)(0)', 'R(1)(1)'] ]
        self.assertEqual(init,c_init)

        print "---- Performing total verification after insert."
        parent.total_verification()

        print "--> removing stars"
        parent.remove_stars()
        print "---- Performing total verification after star removal."
        parent.total_verification()

        print "--> merging"
        parent.merge()
        print "---- Performing total verification after merge."
        parent.total_verification()

        #draw_tree2( parent, 'filename' )


        merge = self.print_tree( parent, "Complex release, merge" )
        c_merge = [ ['A(6)(0)', 'S(2)(0)', 'D(2)(0)', 'K(1)(1)'],\
                    ['A(6)(0)', 'S(2)(0)', 'D(2)(0)', 'R(1)(1)'],\
                    ['A(6)(0)', 'D(4)(2)'],\
                    ['A(6)(0)', 'D(4)(2)', 'K(2)(2)'] ]
        self.assertEqual(merge,c_merge)


        self.assertEqual(A.hidden_labels, { 'D' : { 'S':1, 'T':1 }, 'D,K' : { 'S':1, 'T':1 }  } )

        parent.release_hls(0.2) # hl_release_limit




        res = self.print_tree( parent, "Complex release, final result" ) 
        c_res = [   ['A(6)(0)', 'S(4)(0)', 'D(4)(1)'], 
                    ['A(6)(0)', 'S(4)(0)', 'D(4)(1)', 'K(2)(2)'], 
                    ['A(6)(0)', 'S(4)(0)', 'D(4)(1)', 'R(1)(1)'], 
                    ['A(6)(0)', 'T(2)(0)', 'D(2)(1)'], 
                    ['A(6)(0)', 'T(2)(0)', 'D(2)(1)', 'K(1)(1)']    ]
        self.assertEqual(res,c_res)


    def test_renaming_and_successors( self ) :
        print "\n**** Renaming and successors  ****"
        
        # parent -> *(A,k) -> B -> C -> D 
        # parent -> A -> B -> C -> D
        # parent -> *(A,l) -> *(C,m) -> D

        parent = TreeNode('parent',4,node_count=2)
        A = TreeNode('*',3,node_count=2)
        A.hidden_labels = { 'B,C,D' : { 'A':1, 'k':1 } }
        B = TreeNode('B',2,node_count=2)
        C = TreeNode('C',1,node_count=2)
        D = TreeNode('D',0,node_count=2,pattern_count=2)
        
        A2 = TreeNode('A',3)
        B2 = TreeNode('B',2) 
        C2 = TreeNode('C',1)
        D2 = TreeNode('D',0,pattern_count=1)

        b2 = [ parent, A2, B2, C2, D2 ]
        l2 = [ 0, 1, 0, 1, 0 ]

        # First line:
        parent.set_child(A)
        A.set_child(B)
        B.set_child(C)
        C.set_child(D)
        
        # Second line
        parent.increase_nc()
        parent.set_child(A2)
        A2.set_child(B2)
        B2.set_child(C2)
        C2.set_child(D2)

        # Third line
        parent.increase_nc()
        A2.rename_node('l', b2, l2 )
        B2.increase_nc()
        C2.rename_node('m', b2, l2 )
        D2.increase_nc()
        D2.increase_pc()

        init = self.print_tree( parent, "Renaming and successor" )
        c_init = [ [ '*(4)(0)','B(4)(0)','*(4)(0)','D(4)(4)' ]  ]
        self.assertEqual(init,c_init)

        self.assertEqual(A2.hidden_labels, {'B,D' : { 'A':2, 'k':1, 'l':1 } } )
        self.assertEqual(C2.hidden_labels, {'D' : { 'C':3, 'm':1 } } )


    def test_renaming_and_succcessors2( test ) :
        print "\n**** Renaming and successors2  ****"

        root = TreeNode('root',0,node_count=0)

        lines = [   "Feb 17 00:00:02 <mail.info> eik dovecot: [ID 107833 mail.info] auth(default): shadow(dagolap,129.241.105.215): lookup",\
                    "Feb 17 00:00:02 <mail.info> eik dovecot: [ID 107833 mail.info] auth(default): passwd(dagolap,129.241.105.215): lookup",\
                    "Feb 17 00:00:04 <mail.info> eik dovecot: [ID 107833 mail.info] auth(default): shadow(willa,217.13.20.99): lookup",\
                    "Feb 17 00:00:04 <mail.info> eik dovecot: [ID 107833 mail.info] auth(default): passwd(willa,217.13.20.99): lookup" ]
        
        lines2 = [  "dovecot: ID mail.info auth default shadow dagolap lookup",\
                    "dovecot: ID mail.info auth default passwd dagolap lookup",\
                    "dovecot: ID mail.info auth default shadow willa lookup",\
                    "dovecot: ID mail.info auth default passwd willa lookup" ]
        
        lines3 = [  "auth default shadow dagolap lookup",\
                    "auth default passwd dagolap lookup",\
                    "auth default shadow willa lookup",\
                    "auth default passwd willa lookup" ]
        
        for line in lines :
            msg = extract_message( line )
            #msg = line.split( )
        
            node, line, tail_length, branch, layout = matching_branch = find_matching_branch( root, msg, debug=True )
            add_to_tree( node, line, depth=tail_length, branch=branch, layout=layout )
       
        root.remove_stars()

        print "Performing total verification."  # Note, first after starnodes have been removed!
        root.total_verification()

    def test_renaming_and_successors3( self ) :

        print "\n**** Renaming and successors3  ****"

        parent = TreeNode('parent',5)
        A = TreeNode('A',4)
        B = TreeNode('B',3)
        C = TreeNode('C',2)
        G = TreeNode('G',1)
        F = TreeNode('F',0)
        
        br = [ A, B, C, G, F ]
        l2 = [ 0, 0, 1, 0, 0 ]
        l3 = [ 0, 0, 2, 1, 0 ]
        l4 = [ 0, 0, 2, 2, 0 ]

        # First line
        parent.set_child(A)
        A.set_child(B)
        B.set_child(C)
        C.set_child(G)
        G.set_child(F)
        F.increase_pc()

        # Second line
        parent.increase_nc()
        A.increase_nc()
        B.increase_nc()
        C.rename_node('D',br,l2)
        G.increase_nc()
        F.increase_nc()
        F.increase_pc()

        # Third line
        parent.increase_nc()
        A.increase_nc()
        B.increase_nc()
        C.set_hidden_label('D','G,F','F')
        G.rename_node('W',br,l3)
        F.increase_nc()
        F.increase_pc()

        # Fourth line
        parent.increase_nc()
        A.increase_nc()
        B.increase_nc()
        C.set_hidden_label('D','F')
        G.set_hidden_label('W','F')
        F.increase_nc()
        F.increase_pc()

    
    def test_setting_hidden_labels( self ) :

        # parent -> *(A,B,C)

        parent = TreeNode('parent',1)
        A = TreeNode('A',0)
        
        parent.set_child(A)
        A.increase_pc()

        parent.increase_nc()
        A.rename_node( 'B', [ A ], [ 1 ] )
        A.increase_pc()

        parent.increase_nc()
        A.set_hidden_label( 'C', '' )
        A.increase_pc()

        parent.increase_nc()
        A.set_hidden_label( 'B', '' )
        A.increase_pc()

        parent.increase_nc()
        A.set_hidden_label( 'C', '' )
        A.increase_pc()

        A.verify_hl_count()

        self.assertEqual(A.count, 5)
        self.assertEqual(A.hidden_labels, { '' : { 'A':1, 'B':2, 'C':2 } } )
        

    def test_updating_hidden_labels( self ) :

        star = TreeNode('*', 0, node_count=119, pattern_count=119)
        
        h1 = {  'mbox,data,INBOX,INDEX' : { 'POP3':1, 'IMAP':80 },\
                'Login,user,method,rip,lip,TLS': { 'pop3-login':16, 'imap-login':76 },\
                'mbox,root,index,inbox': { 'POP3':1, 'IMAP':80 } }

        # Renaming from mbox,data,INBOX,INDEX to mbox,INBOX,INDEX
        
        h2 = {  'data,INBOX,INDEX' : { 'POP3':1, 'IMAP':80 },\
                'Login,user,method,rip,lip,TLS': { 'pop3-login':16, 'imap-login':76 },\
                'mbox,root,index,inbox': { 'POP3':1, 'IMAP':80 } }
        
        hl = {  'D,K,G' : { 'A':4, 'B':4, 'C':15, 'D':16 },\
                'P,N,R' : { 'B':12, 'C':11 },\
                'S'     : { 'B':27 },\
                'D,N,S' : { 'C':5, 'D':2 },\
                'D'     : { 'R':23 }  }

        
        star.hidden_labels = hl

        star.update_successor( 'D,K,G', 'D,K', debug=True )

        self.assertEqual(star.sum_hidden_labels(), 119)

        hl2 = { 'D,K' : { 'A':4, 'B':4, 'C':15, 'D':16 },\
                'P,N,R' : { 'B':12, 'C':11 },\
                'S'     : { 'B':27 },\
                'D,N,S' : { 'C':5, 'D':2 },\
                'D'     : { 'R':23 }  }
        
        self.assertEqual(star.hidden_labels, hl2)

    def test_complex_merge( self ) : 

        print "\n**** Complex merge  ****"

        lines2 = [  "Feb 17 14:04:05 <mail.info> pil MSA[12022]: [ID 801593 mail.info] n1HD3aP7012018: to=<sporsmal@sprakradet.no>, delay=00:00:28, xdelay=00:00:28, mailer=esmtp, pri=120659, relay=deframx04.softcom.dk. [80.237.210.74], dsn=2.0.0, stat=Sent (Message queued as: msg.1234875837.50686.22699)",\
                    "Feb 17 14:08:21 <mail.info> pil MSA[12892]: [ID 801593 mail.info] n1HD8I4d012886: to=<jorg@neoplex.org>, delay=00:00:02, xdelay=00:00:00, mailer=esmtp, pri=210882, relay=mail.neoplex.org. [213.179.57.16], dsn=2.0.0, stat=Sent (Ok: queued as 6FE3B807B)",\
                    "Feb 17 14:16:23 <mail.info> pil MSA[13342]: [ID 801593 mail.info] n1HDGJDW013340: to=<oivind_lunde@hotmail.com>, delay=00:00:03, xdelay=00:00:03, mailer=esmtp, pri=193289, relay=mx3.hotmail.com. [65.54.244.200], dsn=2.0.0, stat=Sent ( <20090217131619.GF629@outland.online.ntnu.no> Queued mail for delivery)",\
                    "Feb 17 14:46:04 <mail.info> pil MSA[15066]: [ID 801593 mail.info] n1HDAeK3013065: to=<oyanez@alumnos.inf.utfsm.cl>, delay=00:35:23, xdelay=00:00:07, mailer=esmtp, pri=363523, relay=alegre.inf.utfsm.cl. [204.87.169.3], dsn=2.0.0, stat=Sent (n1HDjvUh008228 Message accepted for delivery)",\
                    "Feb 17 14:46:19 <mail.info> pil MSA[15483]: [ID 801593 mail.info] n1HDkGAb015475: to=<kraakmoen@hotmail.com>, delay=00:00:02, xdelay=00:00:01, mailer=esmtp, pri=126632, relay=mx4.hotmail.com. [65.55.37.104], dsn=2.0.0, stat=Sent ( <20090217144644675.00000002324@adm-171> Queued mail for delivery)"  ]

        root = TreeNode('root',0,node_count=0)

        lines = [   "MSA ID mail.info id1 to delay xdelay mailer esmtp pri relay addr1 dsn stat Sent " + "Message queued as msg1",\
                    "MSA ID mail.info id2 to delay xdelay mailer esmtp pri relay addr2 dsn stat Sent " + "Ok queued as msg2",\
                    "MSA ID mail.info id3 to delay xdelay mailer esmtp pri relay addr3 dsn stat Sent " + "msg3 Queued mail for delivery",\
                    "MSA ID mail.info id4 to delay xdelay mailer esmtp pri relay addr4 dsn stat Sent " + "msg4 Message accepted for delivery ",\
                    "MSA ID mail.info id5 to delay xdelay mailer esmtp pri relay addr5 dsn stat Sent " + "msg5a msg5b Queued mail for delivery " ]

        for line in lines2 :
            msg = extract_message( line )
            #msg = line.split( )
        
            node, line, tail_length, branch, layout = matching_branch = find_matching_branch( root, msg )
            add_to_tree( node, line, depth=tail_length, branch=branch, layout=layout, debug=True )

        #draw_tree2( root, 'filename' )
    
        root.remove_stars()

        #draw_tree2( root, 'filename' )
        
        root.merge(debug=True)

    def test_release_similar( self ) :
        print "**** Release similar: ****"
        # Initial tree:
        # parent -> A -> B -> B

        parent = TreeNode( 'parent', 3, node_count=3)
        A = TreeNode('A', 2, node_count=3)
        A.hidden_labels = {'B':{'B':1}}
        B1 = TreeNode('B', 1, node_count=3, pattern_count=2 )
        B2 = TreeNode('B', 0, node_count=1, pattern_count=1 )

        parent.set_child(A)
        A.set_child(B1)
        B1.set_child(B2)
        
        parent.release_hls( 0.2, debug=True )  # hl_release_limit
        
        # Branch should stay unchanged.
        res = self.print_tree( parent, "Release similar" )
        c_res = [   ['A(3)(0)', 'B(3)(2)'], 
                    ['A(3)(0)', 'B(3)(2)', 'B(1)(1)']   ]
        self.assertEqual(res,c_res)

        



if __name__ == '__main__' :
    unittest.main()
